import collections
import shutil
import sys
import copy
import multiprocessing
import time
import itertools
import argparse
import importlib
import random
import os
import os.path
import types
import uuid
import warnings
import requests
import re
import zipfile
import urllib.request
try:
    import zmq
except:
    zmq = None
from typing import *
try:
    import ujson as json
except:
    import json
try:
    from collections import Iterable
except ImportError:
    from collections.abc import Iterable

import numpy as np
import torch
try:
    import yaml
except ModuleNotFoundError:
    warnings.warn("Module pyyaml is not installed. The configuration cannot be loaded by .yml file.")

import flgo.simulator
import flgo.simulator.default_simulator
import flgo.simulator.base
import flgo.utils.fmodule
import flgo.utils.shared_memory as fus
import flgo.experiment.logger.simple_logger
import flgo.experiment.logger.tune_logger
import flgo.experiment.logger.vertical_logger
import flgo.experiment.logger.dec_logger
import flgo.experiment.logger.hier_logger
import flgo.experiment.logger
import flgo.experiment.logger.pool as felp
import flgo.experiment.device_scheduler
from flgo.simulator.base import BasicSimulator
import flgo.benchmark.base
import flgo.benchmark.partition
import flgo.benchmark.toolkits.partition
import flgo.algorithm

sample_list=['uniform', 'md', 'full', 'uniform_available', 'md_available', 'full_available'] # sampling options for the default sampling method in flgo.algorihtm.fedbase
agg_list=['uniform', 'weighted_scale', 'weighted_com'] # aggregation options for the default aggregating method in flgo.algorihtm.fedbase
optimizer_list=['SGD', 'Adam', 'RMSprop', 'Adagrad'] # supported optimizers
default_option_dict = {'test_parallel':False, 'load_mode': '', 'save_checkpoint':'', 'load_checkpoint':'','pretrain': '', 'sample': 'md', 'aggregate': 'uniform', 'num_rounds': 20, 'proportion': 0.2, 'learning_rate_decay': 0.998, 'lr_scheduler': '-1', 'early_stop': -1, 'num_epochs': 5, 'num_steps': -1, 'learning_rate': 0.1, 'batch_size': 64.0, 'optimizer': 'SGD', 'clip_grad':0.0,'momentum': 0.0, 'weight_decay': 0.0, 'num_edge_rounds':5, 'algo_para': [], 'train_holdout': 0.1, 'test_holdout': 0.0, 'local_test':False, 'local_test_ratio':0.5, 'seed': 0,'dataseed':0, 'gpu': [], 'server_with_cpu': False, 'num_parallels': 1, 'parallel_type':'r', 'num_workers': 0, 'pin_memory':False,'test_batch_size': 512,'pin_memory':False ,'simulator': 'default_simulator', 'availability': 'IDL', 'connectivity': 'IDL', 'completeness': 'IDL', 'responsiveness': 'IDL', 'logger': 'basic_logger', 'log_level': 'INFO', 'log_file': False, 'no_log_console': False, 'no_overwrite': False, 'eval_interval': 1}

if zmq is not None: _ctx = zmq.Context()
else: _ctx = None

class GlobalVariable:
    """This class is to create a shared space for sharing variables across
    different parties for each runner"""

    def __init__(self, logger:flgo.experiment.logger.BasicLogger=None, simulator:flgo.simulator.base.BasicSimulator=None, clock:flgo.simulator.base.ElemClock=None, dev_list:list=None, TaskCalculator:flgo.benchmark.base.BasicTaskCalculator=None, TaskPipe:flgo.benchmark.base.BasicTaskPipe=None):
        self.logger = logger
        self.simulator = simulator
        self.clock = clock
        self.dev_list = dev_list
        self.TaskCalculator = TaskCalculator
        self.TaskPipe = TaskPipe
        self.crt_dev = 0

    def apply_for_device(self):
        r"""'
        Apply for a new device from currently available ones (i.e. devices in self.dev_list)

        Returns:
            GPU device (i.e. torch.device)
        """
        if self.dev_list is None: return None
        dev = self.dev_list[self.crt_dev]
        self.crt_dev = (self.crt_dev + 1) % len(self.dev_list)
        return dev

def setup_seed(seed):
    r"""
    Fix all the random seed used in numpy, torch and random module

    Args:
        seed (int): the random seed
    """
    if seed <0:
        torch.backends.cudnn.enabled = False
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        torch.use_deterministic_algorithms(True)
        seed = -seed
    random.seed(1+seed)
    np.random.seed(21+seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    torch.manual_seed(12+seed)
    torch.cuda.manual_seed_all(123+seed)

def read_option_from_command():
    r"""
    Generate running-time configurations for flgo.init with default values from command lines

    Returns:
        a dict of option (i.e. configuration)
    """

    parser = argparse.ArgumentParser()
    """Training Options"""
    # basic settings
    # methods of server side for sampling and aggregating
    parser.add_argument('--sample', help='methods for sampling clients', type=str, choices=sample_list, default='uniform')
    parser.add_argument('--aggregate', help='methods for aggregating models', type=str, choices=agg_list, default='other')
    # hyper-parameters of training in server side
    parser.add_argument('--num_rounds', help='number of communication rounds', type=int, default=20)
    parser.add_argument('--proportion', help='proportion of clients sampled per round', type=float, default=0.2)
    parser.add_argument('--learning_rate_decay', help='learning rate decay for the training process;', type=float, default=0.998)
    parser.add_argument('--lr_scheduler', help='type of the global learning rate scheduler', type=str, default='-1')
    parser.add_argument('--early_stop', help='stop training if there is no improvement for no smaller than the maximum rounds', type=int, default=-1)
    # hyper-parameters of local_movielens_recommendation training
    parser.add_argument('--num_epochs', help='number of epochs when clients locally train the model on data;', type=int, default=5)
    parser.add_argument('--num_steps', help='the number of local steps, which dominate num_epochs when setting num_steps>0', type=int, default=-1)
    parser.add_argument('--learning_rate', help='learning rate for inner solver;', type=float, default=0.1)
    parser.add_argument('--batch_size', help='batch size', type=float, default='64')
    parser.add_argument('--optimizer', help='select the optimizer for gd', type=str, choices=optimizer_list, default='SGD')
    parser.add_argument('--clip_grad', help='clipping gradients if the max norm of gradients ||g|| > clip_norm > 0', type=float, default=0.0)
    parser.add_argument('--momentum', help='momentum of local training', type=float, default=0.0)
    parser.add_argument('--weight_decay', help='weight decay of local training', type=float, default=0.0)
    parser.add_argument('--num_edge_rounds', help='number of edge rounds in hierFL', type=int, default=5)
    # algorithm-dependent hyper-parameters
    parser.add_argument('--algo_para', help='algorithm-dependent hyper-parameters', nargs='*', type=float)

    """Environment Options"""
    # the ratio of the amount of the data used to train
    parser.add_argument('--train_holdout', help='the rate of holding out the validation dataset from all the local training datasets', type=float, default=0.1)
    parser.add_argument('--test_holdout', help='the rate of holding out the validation dataset from the testing datasets owned by the server', type=float, default=0.0)
    parser.add_argument('--local_test', help='if this term is set True and train_holdout>0, (0.5*train_holdout) of data will be set as client.test_data.', action="store_true", default=False)
    parser.add_argument('--local_test_ratio', help='valid only if local_test==True, the ratio of testing part from train_holdout part', type=float, default=0.5)
    # realistic machine config
    parser.add_argument('--seed', help='seed for random initialization;', type=int, default=0)
    parser.add_argument('--dataseed', help='seed for random initialization for data train/val/test partition', type=int, default=0)
    parser.add_argument('--gpu', nargs='*', help='GPU IDs and empty input is equal to using CPU', type=int)
    parser.add_argument('--server_with_cpu', help='the model parameters will be stored in the memory if True', action="store_true", default=False)
    parser.add_argument('--num_parallels', help="the number of parallels in the clients computing session", type=int, default=1)
    parser.add_argument('--parallel_type', help="the type of parallel: 't' means multi-threading, 'p' means multi-processing and 'r' means ray", type=str, default = 'r')
    parser.add_argument('--num_workers', help='the number of workers of DataLoader', type=int, default=0)
    parser.add_argument('--load_mode', help="the mode of loading task data, e.g., ['', 'mem', 'mmap']", type=str, default='')
    parser.add_argument('--pin_memory', help='pin_memory of DataLoader', action="store_true", default=False)
    parser.add_argument('--drop_last', help='drop_last option of DataLoader, default is False', action="store_true", default=False)
    parser.add_argument('--test_batch_size', help='the batch_size used in testing phase;', type=int, default=512)
    parser.add_argument('--test_parallel', help='whether to use nn.DataParallel to accelerate the testing process', action="store_true", default=False)

    """Simulator Options"""
    # the simulating systemic configuration of clients and the server that helps constructing the heterogeity in the network condition & computing power
    parser.add_argument('--availability', help="client availability mode", type=str, default = 'IDL')
    parser.add_argument('--connectivity', help="client connectivity mode", type=str, default = 'IDL')
    parser.add_argument('--completeness', help="client completeness mode", type=str, default = 'IDL')
    parser.add_argument('--responsiveness', help="client responsiveness mode", type=str, default='IDL')

    """Logger Options"""
    # logger setting
    parser.add_argument('--log_level', help='the level of logger', type=str, default='INFO')
    parser.add_argument('--log_file', help='bool controls whether log to file and default value is False', action="store_true", default=False)
    parser.add_argument('--no_log_console', help='bool controls whether log to screen and default value is True', action="store_true", default=False)
    parser.add_argument('--no_overwrite', help='bool controls whether to overwrite the old result', action="store_true", default=False)
    parser.add_argument('--eval_interval', help='evaluate every __ rounds;', type=int, default=1)
    parser.add_argument('--save_checkpoint', help='the level of logger', type=str, default='')
    parser.add_argument('--load_checkpoint', help='the level of logger', type=str, default='')

    try: option = vars(parser.parse_known_args()[0])
    except IOError as msg: parser.error(str(msg))
    for key in option.keys():
        if option[key] is None:
            option[key]=[]
    return option

def option_helper():
    from flgo.utils import option_desc
    import prettytable as pt
    lines = option_desc.split('\n')
    lines = [l.split(',') for l in lines if len(l)>0]
    tb = pt.PrettyTable(lines[0])
    for i in range(1, len(lines)):
        tb.add_row(lines[i])
    print(tb)
    return

def load_configuration(config={}):
    r"""
    Load configurations from .yml file or dict.

    Args:
        config (dict|str): the configurations

    Returns:
        a dict of option (i.e. configuration)
    """
    if type(config) is str and config.endswith('.yml'):
        with open(config) as f:
            option = yaml.load(f, Loader=yaml.FullLoader)
        return option
    elif type(config) is dict:
        return config
    else:
        raise TypeError('The input config should be either a dict or a filename.')

def gen_benchmark_from_file(benchmark:str, config_file:str, target_path='.',data_type:str='cv', task_type:str='classification', overwrite:bool=False) -> str:
    r"""
        Create customized benchmarks from configurations. The configuration is a .py file that describes the datasets and the model,
        where there must exist a function named `get_model` and a variable `train_data`. `val_data` and test_data are two optional
        variables in the configuration.
    Args:
        benchmark (str): the name of the benchmark
        config_file (str): the path of the configuration file
        target_path: (str): the path to store the benchmark
        data_type (str): the type of dataset that should be in the list ['cv', 'nlp', 'graph', 'rec', 'series', 'tabular']
        task_type (str): the type of the task (e.g. classification, regression...)
        overwrite (bool): overwrite current benchmark if there already exists a benchmark of the same name
    Returns:
        bmk_module (str): the module name of the generated benchmark
    """
    if not os.path.exists(config_file): raise FileNotFoundError('File {} not found.'.format(config_file))
    target_path = os.path.abspath(target_path)
    bmk_path = os.path.join(target_path, benchmark)
    if os.path.exists(bmk_path):
        if not overwrite:
            warnings.warn('There already exists a benchmark `{}`'.format(benchmark))
            return '.'.join(os.path.relpath(bmk_path, os.getcwd()).split(os.path.sep))
        # raise FileExistsError('Benchmark {} already exists'.format(bmk_path))
    if data_type=='general':
        temp_path = os.path.join(flgo.benchmark.path, 'toolkits', 'general')
    else:
        temp_path = os.path.join(flgo.benchmark.path, 'toolkits', data_type, task_type, 'temp')
    if not os.path.exists(temp_path):
        raise NotImplementedError('There is no support to automatically generation of {}.{}. More other types are comming soon...'.format(data_type, task_type))
    else:
        shutil.copytree(temp_path, bmk_path)
    shutil.copyfile(config_file, os.path.join(bmk_path, 'config.py'))
    bmk_module = '.'.join(os.path.relpath(bmk_path, os.getcwd()).split(os.path.sep))
    return bmk_module

def gen_benchmark(benchmark:str, config_file:str, target_path='.',data_type:str='cv', task_type:str='classification'):
    r"""
        Create customized benchmarks from configurations. The configuration is a .py file that describes the datasets and the model,
        where there must exist a function named `get_model` and a variable `train_data`. `val_data` and test_data are two optional
        variables in the configuration.
    Args:
        benchmark (str): the name of the benchmark
        config_file (str): the path of the configuration file
        target_path: (str): the path to store the benchmark
        data_type (str): the type of dataset that should be in the list ['cv', 'nlp', 'graph', 'rec', 'series', 'tabular']
        task_type (str): the type of the task (e.g. classification, regression...)
    Returns:
        bmk_module (str): the module name of the generated benchmark
    """
    if not os.path.exists(config_file): raise FileNotFoundError('File {} not found.'.format(config_file))
    target_path = os.path.abspath(target_path)
    bmk_path = os.path.join(target_path, benchmark)
    if os.path.exists(bmk_path): raise FileExistsError('Benchmark {} already exists'.format(bmk_path))
    if data_type=='general':
        temp_path = os.path.join(flgo.benchmark.path, 'toolkits', 'general')
    else:
        temp_path = os.path.join(flgo.benchmark.path, 'toolkits', data_type, task_type, 'temp')
    if not os.path.exists(temp_path):
        raise NotImplementedError('There is no support to automatically generation of {}.{}. More other types are comming soon...'.format(data_type, task_type))
    else:
        shutil.copytree(temp_path, bmk_path)
    shutil.copyfile(config_file, os.path.join(bmk_path, 'config.py'))
    bmk_module = '.'.join(os.path.relpath(bmk_path, os.getcwd()).split(os.path.sep))
    return bmk_module

def gen_decentralized_benchmark(benchmark:str, config_file:str, target_path = '.', data_type:str='cv', task_type:str='classification'):
    r"""
        Create customized benchmarks from configurations. The configuration is a .py file that describes the datasets and the model,
        where there must exist a function named `get_model` and a variable `train_data`. `val_data` and test_data are two optional
        variables in the configuration.
    Args:
        benchmark (str): the name of the benchmarkc
        config_file (str): the path of the configuration file
        target_path: (str): the path to store the benchmark
        data_type (str): the type of dataset that should be in the list ['cv', 'nlp', 'graph', 'rec', 'series', 'tabular']
        task_type (str): the type of the task (e.g. classification, regression...)
    Returns:
        bmk_module (str): the module name of the generated benchmark
    """
    if not os.path.exists(config_file): raise FileNotFoundError('File {} not found.'.format(config_file))
    target_path = os.path.abspath(target_path)
    bmk_path = os.path.join(target_path, benchmark)
    if os.path.exists(bmk_path): raise FileExistsError('Benchmark {} already exists'.format(bmk_path))
    temp_path = os.path.join(flgo.benchmark.path, 'toolkits', data_type, task_type, 'dec_temp')
    if not os.path.exists(temp_path):
        raise NotImplementedError('There is no support to automatically generation of {}.{}. More other types are comming soon...'.format(data_type, task_type))
    else:
        shutil.copytree(temp_path, bmk_path)
    shutil.copyfile(config_file, os.path.join(bmk_path, 'config.py'))
    bmk_module = '.'.join(os.path.relpath(bmk_path, os.getcwd()).split(os.path.sep))
    return bmk_module

def gen_hierarchical_benchmark(benchmark:str, config_file:str, target_path = '.', data_type:str='cv', task_type:str='classification'):
    r"""
        Create customized benchmarks from configurations. The configuration is a .py file that describes the datasets and the model,
        where there must exist a function named `get_model` and a variable `train_data`. `val_data` and test_data are two optional
        variables in the configuration.
    Args:
        benchmark (str): the name of the benchmark
        config_file (str): the path of the configuration file
        target_path: (str): the path to store the benchmark
        data_type (str): the type of dataset that should be in the list ['cv', 'nlp', 'graph', 'rec', 'series', 'tabular']
        task_type (str): the type of the task (e.g. classification, regression...)
    Returns:
        bmk_module (str): the module name of the generated benchmark
    """
    if not os.path.exists(config_file): raise FileNotFoundError('File {} not found.'.format(config_file))
    target_path = os.path.abspath(target_path)
    bmk_path = os.path.join(target_path, benchmark)
    if os.path.exists(bmk_path): raise FileExistsError('Benchmark {} already exists'.format(bmk_path))
    temp_path = os.path.join(flgo.benchmark.path, 'toolkits', data_type, task_type, 'hier_temp')
    if not os.path.exists(temp_path):
        raise NotImplementedError('There is no support to automatically generation of {}.{}. More other types are comming soon...'.format(data_type, task_type))
    else:
        shutil.copytree(temp_path, bmk_path)
    shutil.copyfile(config_file, os.path.join(bmk_path, 'config.py'))
    bmk_module = '.'.join(os.path.relpath(bmk_path, os.getcwd()).split(os.path.sep))
    return bmk_module

def gen_real_task(benchmark, config_file:str, task_path:str='', seed=0, overwrite=False):
    """
       Generate federated task from benchmark and partitioner without inputing other parameters
       Args:
           benchmark (module): benchmark
           partitioner (flgo.benchmark.partition.BasicPartitioner): a instance of type flgo.benchmark.partition.BasicPartitioner
           task_path (str): the name and the path of the task
           seed (int): random seed
           overwrite (bool): overwrite the old task if the task_path already exist if True

       Returns:
           task_path (str): the path of the task
       """
    # generate the name of task randomly if empty
    if len(task_path) == 0: task_path = 'FLGoTask_' + uuid.uuid4().hex
    # setup random seed
    random.seed(3 + seed)
    np.random.seed(97 + seed)
    torch.manual_seed(12 + seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    bmk_core = importlib.import_module('.core', benchmark.__name__)
    # load configuration
    task_generator = getattr(bmk_core, 'TaskGenerator')()
    # check if task already exists
    task_pipe = getattr(bmk_core, 'TaskPipe')(task_path)
    if task_pipe.task_exists():
        if not overwrite:
            warnings.warn(
                'Task {} already exists. To overwrite the existing task, use flgo.gen_task_by_(...,overwrite=True,...)'.format(
                    task_path))
            return
        else:
            shutil.rmtree(task_path)
    # save the generated federated benchmark
    # initialize task pipe
    try:
        # create task architecture
        task_pipe.create_task_architecture()
        # save meta infomation
        task_pipe.save_info(task_generator)
        shutil.copyfile(config_file, os.path.join(task_path, 'dataset.py'))
        print('Task {} has been successfully generated.'.format(task_pipe.task_path))
    except Exception as e:
        print(e)
        task_pipe.remove_task()
        print("Failed to saving task.")
        return None
    return task_path

def gen_task(config={}, task_path:str= '', rawdata_path:str= '', seed:int=0, overwrite:bool=False):
    r"""
    Generate a federated task that is specified by the benchmark information and the partition information, where the generated task will be stored in the task_path and the raw data will be downloaded into the rawdata_path.

        config (dict || str): configuration is either a dict contains parameters or a filename of a .yml file
        task_path (str): where the generated task will be stored
        rawdata_path (str): where the raw data will be downloaded\stored
        seed (int): the random seed used to generate the task

    Example:
    ```python
        >>> import flgo
        >>> config = {'benchmark':{'name':'flgo.benchmark.mnist_classification'}, 'partitioner':{'name':'IIDParitioner', 'para':{'num_clients':100}}}
        >>> flgo.gen_task(config, './my_mnist_iid')
        >>> # The task will be stored as `my_mnist_iid` in the current working dictionary
    ```
    """
    # setup random seed
    random.seed(3 + seed)
    np.random.seed(97 + seed)
    torch.manual_seed(12+seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    # load configuration
    gen_option = load_configuration(config)
    if type(gen_option['benchmark']) is not dict: gen_option['benchmark']={'name':gen_option['benchmark']}
    if 'para' not in gen_option['benchmark'].keys(): gen_option['benchmark']['para'] = {}
    # init generator
    if rawdata_path!='': gen_option['benchmark']['para']['rawdata_path']=rawdata_path
    if type(gen_option['benchmark']['name']) is str:
        bmk_core = importlib.import_module('.'.join([gen_option['benchmark']['name'], 'core']))
    elif hasattr(gen_option['benchmark']['name'], '__path__'):
        bmk_core = importlib.import_module('.core', gen_option['benchmark']['name'].__name__)
    else:
        raise RuntimeError("The value of parameter config['benchmark']['name'] should be either a string or a python package.")
    task_generator = getattr(bmk_core, 'TaskGenerator')(**gen_option['benchmark']['para'])
    bmk_module = importlib.import_module(gen_option['benchmark']['name']) if type(
        gen_option['benchmark']['name']) is str else gen_option['benchmark']['name']
    # create partitioner for generator if specified
    if 'partitioner' in gen_option.keys():
        if isinstance(gen_option['partitioner'], flgo.benchmark.partition.BasicPartitioner):
            partitioner = gen_option['partitioner']
            task_generator.register_partitioner(partitioner)
            partitioner.register_generator(task_generator)
        else:
            if not isinstance(gen_option['partitioner'], dict):
                gen_option['partitioner'] = {'name': gen_option['partitioner'], 'para':{}}
            # update parameters of partitioner
            if 'para' not in gen_option['partitioner'].keys():
                gen_option['partitioner']['para'] = {}
            else:
                if 'name' not in gen_option['partitioner'].keys():
                    gen_option['benchmark']['para'].update(gen_option['partitioner']['para'])
            if 'name' in gen_option['partitioner'].keys():
                Partitioner = gen_option['partitioner']['name']
                if type(Partitioner) is str:
                    if Partitioner in globals().keys(): Partitioner = eval(Partitioner)
                    else: Partitioner = getattr(flgo.benchmark.partition, Partitioner)
                partitioner = Partitioner(**gen_option['partitioner']['para'])
                task_generator.register_partitioner(partitioner)
                partitioner.register_generator(task_generator)
            else:
                try:
                    if hasattr(bmk_module, 'default_partitioner'):
                        Partitioner = getattr(bmk_module, 'default_partitioner')
                        default_partition_para = getattr(bmk_module, 'default_partition_para') if hasattr(bmk_module, 'default_partition_para') else {}
                        partitioner = Partitioner(**default_partition_para)
                        task_generator.register_partitioner(partitioner)
                        partitioner.register_generator(task_generator)
                    else:
                        partitioner = None
                except:
                    partitioner = None
    # initialize task pipe
    if len(task_path) == 0: task_path = 'FLGoTask_' + uuid.uuid4().hex
    task_pipe = getattr(bmk_core, 'TaskPipe')(task_path)
    # check if task already exists
    if task_pipe.task_exists():
        if not overwrite:
            warnings.warn('Task {} already exists. To overwrite the existing task, use flgo.gen_task(...,overwrite=True,...)'.format(task_path))
            return
        else:
            shutil.rmtree(task_path)
    # generate federated task
    task_generator.generate()
    # save the generated federated benchmark
    try:
        # create task architecture
        task_pipe.create_task_architecture()
        # save meta infomation
        task_pipe.save_info(task_generator)
        # save task
        task_pipe.save_task(task_generator)
        print('Task {} has been successfully generated.'.format(task_pipe.task_path))
    except Exception as e:
        print(e)
        task_pipe.remove_task()
        print("Failed to saving splited dataset.")
        return None
    # save visualization
    try:
        visualize_func = getattr(bmk_module,'visualize')
        visualize_func(task_generator, partitioner, task_path)
    except Exception as e:
        print('Warning: Failed to visualize the partitioned result where there exists error {}'.format(e))
    finally:
        return task_path

def gen_task_by_(benchmark, partitioner:flgo.benchmark.partition.BasicPartitioner=None, task_path:str='', seed:int=0, overwrite:bool=False):
    """
    Generate federated task from benchmark and partitioner without inputing other parameters
    Args:
        benchmark (module): benchmark
        partitioner (flgo.benchmark.partition.BasicPartitioner): a instance of type flgo.benchmark.partition.BasicPartitioner
        task_path (str): the name and the path of the task
        seed (int): random seed
        overwrite (bool): overwrite the old task if the task_path already exist if True

    Returns:
        task_path (str): the path of the task
    """
    # generate the name of task randomly if empty
    if len(task_path)==0: task_path = 'FLGoTask_'+uuid.uuid4().hex
    # setup random seed
    random.seed(3 + seed)
    np.random.seed(97 + seed)
    torch.manual_seed(12 + seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    bmk_core = importlib.import_module('.core', benchmark.__name__)
    # load configuration
    task_generator = getattr(bmk_core, 'TaskGenerator')()
    if partitioner is not None:
        task_generator.register_partitioner(partitioner)
        partitioner.register_generator(task_generator)
    # check if task already exists
    task_pipe = getattr(bmk_core, 'TaskPipe')(task_path)
    if task_pipe.task_exists():
        if not overwrite:
            warnings.warn('Task {} already exists. To overwrite the existing task, use flgo.gen_task_by_(...,overwrite=True,...)'.format(task_path))
            return
        else:
            shutil.rmtree(task_path)
    # generate federated task
    task_generator.generate()
    # save the generated federated benchmark
    # initialize task pipe
    try:
        # create task architecture
        task_pipe.create_task_architecture()
        # save meta infomation
        task_pipe.save_info(task_generator)
        # save task
        task_pipe.save_task(task_generator)
        print('Task {} has been successfully generated.'.format(task_pipe.task_path))
    except Exception as e:
        print(e)
        task_pipe.remove_task()
        print("Failed to saving splited dataset.")
        return None
    # save visualization
    try:
        visualize_func = getattr(benchmark, 'visualize')
        visualize_func(task_generator, partitioner, task_path)
    except Exception as e:
        print('Warning: Failed to visualize the partitioned result where there exists error {}'.format(e))
    finally:
        return task_path

def gen_direct_task_from_file(task: str, config_file: str, target_path='.', overwrite:bool=False):
    r"""

    Args:
        task (str): the name of the task
        config_file (str): the path of the configuration file
        target_path: (str): the path to store the benchmark
        overwrite (bool): overwrite current benchmark if there already exists a benchmark of the same name
    Returns:
        task (str): the task name
    """
    if not os.path.exists(config_file): raise FileNotFoundError('File {} not found.'.format(config_file))
    target_path = os.path.abspath(target_path)
    bmk_path = os.path.join(target_path, task)
    if os.path.exists(bmk_path):
        if not overwrite:
            warnings.warn('There already exists a task `{}`'.format(task))
            return '.'.join(os.path.relpath(bmk_path, os.getcwd()).split(os.path.sep))
        else:
            shutil.rmtree(bmk_path)
    temp_path = os.path.join(flgo.benchmark.path, 'toolkits', 'direct')
    shutil.copytree(temp_path, bmk_path)
    shutil.copyfile(config_file, os.path.join(bmk_path, 'config.py'))
    bmk_module = '.'.join(os.path.relpath(bmk_path, os.getcwd()).split(os.path.sep))
    bmk_config = importlib.import_module('.'.join([bmk_module, 'config']))
    train_data = getattr(bmk_config, 'train_data')
    # generate federated task
    # save the generated federated benchmark
    # initialize task pipe
    try:
        # create task architecture
        os.mkdir(os.path.join(bmk_path, 'record'))
        os.mkdir(os.path.join(bmk_path, 'log'))
        info = {'benchmark': bmk_module, 'bmk_path': bmk_path}
        info['scene'] = 'horizontal'
        info['num_clients'] = 'unknown'
        with open(os.path.join(bmk_path, 'info'), 'w') as outf:
            json.dump(info, outf)
        data_json = {"client_names": [f"Client{cid}" for cid in range(len(train_data))]}
        with open(os.path.join(bmk_path, 'data.json'), 'w') as outf2:
            json.dump(data_json, outf2)
        print('Task {} has been successfully generated.'.format(bmk_path))
    except Exception as e:
        print(e)
        shutil.rmtree(bmk_path)
        print("Failed to saving splited dataset.")
        return None
    # save visualization
    return bmk_path

def load_task_data(task, train_holdout:float=0.2, test_holdout:float=0.0, local_test=False, local_test_ratio=0.5, seed=0):
    r"""
    Load task datasets from a given task and some configurations.

    Args:
        task (str): the dictionary of the federated task
        train_holdout (float): the ratio of holdout local data for each client (e.g., local validation data)
        test_holdout (float): the ratio of holdout local data for the server (e.g., global validation data or public data)
        local_test (bool): whether to split the holdout training data again
        local_test_ratio (float): the ratio of twice-holdout local data (e.g., local testing data)
        seed (int): random seed for reproducibility

    Returns:
        task_data (dict): a dict consists of datasets for different parties in FL.
    """
    if not os.path.exists(task):
        raise FileExistsError("Fedtask '{}' doesn't exist. Please generate the specified task by flgo.gen_task().")
    with open(os.path.join(task, 'info'), 'r') as inf:
        task_info = json.load(inf)
    benchmark = task_info['benchmark']
    if 'bmk_path' in task_info and os.path.isdir(task_info['bmk_path']): sys.path.append(task_info['bmk_path'])
    core_module = '.'.join([benchmark, 'core'])
    task_pipe = getattr(importlib.import_module(core_module), 'TaskPipe')(task)
    setup_seed(seed)
    option = {'train_holdout':train_holdout, 'test_holdout':test_holdout, 'local_test':local_test, 'local_test_ratio':local_test_ratio}
    task_data = task_pipe.load_data(option)
    return task_data

def init(task: str, algorithm, option = {}, model=None, Logger: flgo.experiment.logger.BasicLogger = None, Simulator: BasicSimulator=flgo.simulator.DefaultSimulator, scene='horizontal'):
    r"""
    Initialize a runner in FLGo, which is to optimize a model on a specific task (i.e. IID-mnist-of-100-clients) by the selected federated algorithm.

    Args:
        task (str): the dictionary of the federated task
        algorithm (module|class): the algorithm will be used to optimize the model in federated manner, which must contain pre-defined attributions (e.g. algorithm.Server and algorithm.Client for horizontal federated learning)
        option (dict|str): the configurations of training, environment, algorithm, logger and simulator
        model (module|class): the model module that contains two methods: model.init_local_module(object) and model.init_global_module(object)
        Logger (flgo.experiment.logger.BasicLogger): the class of the logger inherited from flgo.experiment.logger.BasicLogger
        Simulator (flgo.simulator.base.BasicSimulator): the class of the simulator inherited from flgo.simulator.BasicSimulator
        scene (str): 'horizontal' or 'vertical' in current version of FLGo

    Returns:
        runner: the object instance that has the method object.run()

    Option:
        sample              (str):  methods for sampling client, which should be one of 'uniform' (default), 'md', 'full' or each with an additional suffix '_available', e.g., 'full_available'
        aggregate           (str):  methods for aggregating models,  which should be one of 'other' (default standard fedavg), 'uniform', 'weighted_scale', 'weighted_com'
        num_rounds          (int):  number of communication rounds, default is 20
        proportion          (flt):  proportion of clients sampled per round, default=0.2
        learning_rate_decay (flt):  learning rate decay for the training process, default=0.998
        lr_scheduler        (str):  type of the global learning rate scheduler, default='-1' (no decay), 0 is round-wise decay lr = lr*decay^round
        early_stop          (int):  stop training if there is no improvement for no smaller than the maximum rounds, default=-1 (no early stop)
        num_epochs          (int):  number of epochs when clients locally train the model on data, default=5
        num_step            (int):  the number of local steps, which dominate num_epochs when setting num_steps>0', default=-1
        learning_rate       (flt):  learning rate for inner solver;', type=float, default=0.1)
        batch_size          (int):  batch size', type=float, default='64')
        optimizer           (str):  select the optimizer for gd', type=str, choices=optimizer_list, default='SGD')
        clip_grad           (flt):  clipping gradients if the max norm of gradients ||g|| > clip_norm > 0, default=0.0
        momentum            (flt):  momentum of local training, default=0.0
        weight_decay        (flt):  weight decay of local training',default=0.0
        num_edge_rounds     (int):  number of edge rounds in hierFL',default=5
        algo_para           (any):  algorithm-dependent hyper-parameters
        train_holdout       (flt):  the rate of holding out the validation dataset from all the local training datasets',default=0.1
        test_holdout        (flt):  the rate of holding out the validation dataset from the testing datasets owned by the server', default=0.0
        local_test          (bool): if this term is set True and train_holdout>0, (0.5*train_holdout) of data will be set as client.test_data.default=False.
        local_test_ratio    (flt):  valid if local_test is True. The ratio of local test dataset holdout from local validation data.
        seed                (flt):  seed for random initialization;', default=0
        dataseed            (int):  seed for random initialization for data train/val/test partition', default=0
        gpu                 (list\int): GPU IDs and empty input is equal to using CPU',
        server_with_cpu     (bool): the model parameters will be stored in the memory if True, default=False
        num_parallels       (int):  the number of parallels in the clients computing session, default=1
        parallel_type       (str):  the type of parallel: 't' means multi-threading and 'p' means multi-processing, default = 't'
        num_workers         (int):  the number of workers of DataLoader, default=0
        load_mode           (str):  the loading mode of datasets
        pin_memory          (bool): pin_memory of DataLoader, default=False
        drop_last           (bool): drop_last option of DataLoader, default is False
        test_batch_size     (int):  the batch_size used in testing phase, default=512
        availability        (str):  client availability mode, default = 'IDL'
        connectivity        (str):  client connectivity mode, default = 'IDL'
        completeness        (str):  client completeness mode, default = 'IDL'
        responsiveness      (str):  client responsiveness mode, default='IDL'
        log_level           (str):  the level of logger, default='INFO'
        log_file            (bool): whether log to file and default value is False, default=False
        no_log_console      (bool): whether log to screen and default value is True, default=False
        no_overwrite        (bool): whether to overwrite the old result, default=False
        eval_interval       (int):  evaluate every __ rounds, default=1
        save_checkpoint     (str):  the level of logger, default=''
        load_checkpoint     (str):  the level of logger, default=''

    Example:
    ```python
        >>> import flgo
        >>> from flgo.algorithm import fedavg
        >>> from flgo.experiment.logger.simple_logger import SimpleLogger
        >>> # create task 'mnist_iid' by flgo.gen_task('gen_config.yml', 'mnist_iid') if there exists no such task
        >>> if os.path.exists('mnist_iid'): flgo.gen_task({'benchmark':{'name':'flgo.benchmark.mnist_classification'}, 'partitioner':{'name':'IIDPartitioner','para':{'num_clients':100}}}, 'mnist_iid')
        >>> # create runner
        >>> fedavg_runner = flgo.init('mnist_iid', algorithm=fedavg, option = {'num_rounds':20, 'gpu':[0], 'learning_rate':0.1})
        >>> fedavg_runner.run()
        ... # the training will start after runner.run() was called, and the running-time results will be recorded by Logger into the task dictionary
    ```
    """
    # init option
    option = load_configuration(option)
    default_option = read_option_from_command()
    for op_key in option:
        if op_key in default_option.keys():
            op_type = type(default_option[op_key])
            if op_type == type(option[op_key]):
                default_option[op_key] = option[op_key]
            else:
                if op_type is list:
                    default_option[op_key]=list(option[op_key]) if hasattr(option[op_key], '__iter__') else [option[op_key]]
                elif op_type is tuple:
                    default_option[op_key] = tuple(option[op_key]) if hasattr(option[op_key], '__iter__') else (option[op_key])
                else:
                    default_option[op_key] = op_type(option[op_key])
        else:
            default_option[op_key] = option[op_key]
    option = default_option
    setup_seed(seed=option['seed'])
    option['task'] = task
    option['algorithm'] = (algorithm.__name__).split('.')[-1]
    option['scene'] = scene
    option['simulator'] = Simulator.__name__ if Simulator is not None else 'None'
    # # init devices
    # if not (option['gpu'] is None or len(option['gpu']) == 0) and option['gpu'][0]>0:
    #     os.environ['CUDA_VISIBLE_DEVICES'] = ','.join([str(i) for i in option['gpu']])
    #     option['gpu'] = list(range(len(option['gpu'])))
    # init task information
    if not os.path.exists(task):
        raise FileExistsError("Fedtask '{}' doesn't exist. Please generate the specified task by flgo.gen_task().")
    with open(os.path.join(task, 'info'), 'r') as inf:
        task_info = json.load(inf)
    # benchmark information
    benchmark = task_info['benchmark']
    if 'bmk_path' in task_info and os.path.isdir(task_info['bmk_path']): sys.path.append(task_info['bmk_path'])
    # model information
    if model== None:
        bmk_module = importlib.import_module(benchmark)
        if hasattr(algorithm, 'init_global_module') or hasattr(algorithm, 'init_local_module'):
            model = algorithm
        elif hasattr(bmk_module, 'default_model'):
            model = getattr(bmk_module, 'default_model')
        else:
            raise NotImplementedError("Model cannot be None when there exists no default model for the current benchmark {} and the algorithm {} didn't define the model by `init_local_module` or `init_global_module`".format(task_info['benchmark'], option['algorithm']))
    option['model'] = (model.__name__).split('.')[-1]

    # create global variable
    gv = GlobalVariable()
    # init logger
    default_scene_logger = {
        'horizontal': felp.SimpleLogger,
        'vertical': felp.VerticalLogger,
        'real_hclient': felp.ParallelHFLLogger,
        'real_hserver': felp.ParallelHFLLogger,
        'parallel_horizontal': felp.ParallelHFLLogger,
        'decentralized': felp.DecLogger,
        'hierarchical':felp.HierLogger,
    }
    assert scene in default_scene_logger.keys()
    if Logger is None: Logger = default_scene_logger[scene]
    logger = Logger(task=task, option=option, name=str(id(gv))+str(Logger), level=option['log_level'])
    gv.logger = logger

    # init device
    if (option['gpu'] is None or len(option['gpu']) == 0):
        gv.dev_list = [torch.device('cpu')]
    else:
        if int(option['gpu'][0])<0: # for mac equiped with M1
            gv.dev_list = [torch.device('mps')]
        else: # for other machines equiped with GPU
            gv.dev_list = [torch.device('cuda:{}'.format(gpu_id)) for gpu_id in option['gpu']]
    logger.info('PROCESS ID:\t{}'.format(os.getpid()))
    logger.info('Initializing devices: '+','.join([str(dev) for dev in gv.dev_list])+' will be used for this running.')

    # init task
    logger.info('BENCHMARK:\t{}'.format(benchmark))
    logger.info('TASK:\t\t\t{}'.format(task))
    logger.info('MODEL:\t\t{}'.format(model.__name__))
    logger.info('ALGORITHM:\t{}'.format(option['algorithm']))
    core_module = '.'.join([benchmark, 'core'])
    gv.TaskPipe = getattr(importlib.import_module(core_module), 'TaskPipe')
    task_pipe = gv.TaskPipe(task)
    TaskCalculator = getattr(importlib.import_module(core_module), 'TaskCalculator')
    gv.TaskCalculator = TaskCalculator
    setup_seed(option['dataseed'])

    # scene-specific procedure
    ############################################### Simulation FL ###########################################
    if scene in ['horizontal', 'vertical', 'decentralized', 'hierarchical','parallel_horizontal']:
        if option['load_mode']=='mmap':
            task_data = flgo.utils.shared_memory.load_task_data_from_npz(task, train_holdout=option['train_holdout'], test_holdout=option['test_holdout'], local_test=option['local_test'], local_test_ratio=option['local_test_ratio'], seed=option['dataseed'], create=True)
        else:
            task_data = task_pipe.load_data(option)
            if option['load_mode']=='mem':
                for u in task_data:
                    for v in task_data[u]:
                        task_data[u][v] = flgo.utils.shared_memory.MemDataset(task_data[u][v]) if task_data[u][v] is not None else None
        # init objects
        obj_class = [c for c in dir(algorithm) if not c.startswith('__')]
        tmp = []
        for c in obj_class:
            try:
                C = getattr(algorithm, c)
                setattr(C, 'gv', gv)
                setattr(C, 'TaskCalculator', TaskCalculator)
                tmp.append(c)
            except:
                continue
        if scene == 'horizontal':
            for c in obj_class:
                if 'Client' in c:
                    class_client = getattr(algorithm, c)
                    class_client.train = flgo.simulator.base.with_completeness(class_client.train)
                elif 'Server' in c:
                    class_server = getattr(algorithm, c)
                    class_server.sample = flgo.simulator.base.with_availability(class_server.sample)
                    class_server.communicate_with = flgo.simulator.base.with_latency(class_server.communicate_with)
                    class_server.communicate = flgo.simulator.base.with_dropout(class_server.communicate)
        objects = task_pipe.generate_objects(option, algorithm, scene=scene)
        obj_classes = collections.defaultdict(int)
        for obj in objects: obj_classes[obj.__class__] += 1
        creating_str = []
        for k, v in obj_classes.items(): creating_str.append("{} {}".format(v, k))
        creating_str = ', '.join(creating_str)
        logger.info('SCENE:\t\t{} FL with '.format(scene) + creating_str)
        task_pipe.distribute(task_data, objects)

        # init model
        if hasattr(model, 'init_local_module'):
            for object in objects:
                model.init_local_module(object)
        if hasattr(model, 'init_global_module'):
            for object in objects:
                model.init_global_module(object)
        if hasattr(model, 'init_dataset'):
            for object in objects:
                model.init_dataset(object)
        setup_seed(option['seed'] + 346)

        # init communicator
        gv.communicator = flgo.VirtualCommunicator(objects)
        logger.info('SIMULATOR:\t{}'.format(str(Simulator)))
        gv.clock = flgo.simulator.base.ElemClock()
        gv.simulator = Simulator(objects, option) if scene == 'horizontal' else None
        if gv.simulator is not None: gv.simulator.initialize()
        gv.clock.register_simulator(simulator=gv.simulator)

        # final initialize
        for ob in objects: ob.initialize()

        # init virtual system environment
        logger.register_variable(coordinator=objects[0], participants=objects[1:], option=option, clock=gv.clock, scene=scene, objects=objects, simulator=Simulator.__name__ if scene == 'horizontal' else 'None')
        if scene == 'horizontal': logger.register_variable(server=objects[0], clients=objects[1:])
        logger.initialize()
        logger.info('Ready to start.')
        # register global variables for objects
        for c in tmp:
            try:
                C = getattr(algorithm, c)
                delattr(C, 'gv')
            except:
                continue
        for ob in objects:
            ob.gv = gv
        if gv.simulator is not None:
            gv.simulator.gv = gv
        gv.clock.gv = gv
        logger.gv = gv
        if scene=='parallel_horizontal':
            for obj in objects: obj.logger = Logger
            class HParallelRunner:
                def __init__(self, objects:list):
                    self.objects = objects
                    self.plist = []

                def run(self):
                    for obj in self.objects:
                        p = torch.multiprocessing.Process(target=obj.run, )
                        p.start()
                        self.plist.append(p)
            objects = [HParallelRunner(objects)]
    ############################################### Real World FL #####################################################
    else:
        # load task data
        sys.path.append(task)
        task_config = importlib.import_module('dataset')
        train_data = getattr(task_config, 'train_data') if hasattr(task_config, 'train_data') else None
        test_data = getattr(task_config, 'test_data') if hasattr(task_config, 'test_data') else None
        val_data = getattr(task_config, 'val_data') if hasattr(task_config, 'val_data') else None
        task_data = {'train':train_data, 'test':test_data, 'val':val_data}

        # init objects
        obj_class = [c for c in dir(algorithm) if not c.startswith('__')]
        tmp = []
        for c in obj_class:
            try:
                C = getattr(algorithm, c)
                setattr(C, 'gv', gv)
                setattr(C, 'TaskCalculator', TaskCalculator)
                tmp.append(c)
            except:
                continue
        # distribute local data to objects
        objects = task_pipe.generate_objects(option, algorithm, scene=scene)
        if scene=='real_hclient':
            # load self name in the current task
            name_path = os.path.join(task, 'name')
            if not os.path.exists(name_path):
                name = _get_name()
                with open(name_path, 'w') as namefile:
                    namefile.write(name)
            with open(name_path, 'r') as namefile:
                objects[0].name = namefile.readline()
        obj_classes = collections.defaultdict(int)
        for obj in objects: obj_classes[obj.__class__]+=1
        creating_str = []
        for k,v in obj_classes.items(): creating_str.append("{} {}".format(v, k))
        creating_str = ', '.join(creating_str)
        logger.info('SCENE:\t\t{} FL with '.format(scene)+creating_str)
        task_data = {objects[0].name: task_data}
        objects[0].logger = Logger
        task_pipe.distribute(task_data, objects)
        # init model
        if hasattr(model, 'init_local_module'):
            for object in objects: model.init_local_module(object)
        if hasattr(model, 'init_global_module'):
            for object in objects: model.init_global_module(object)
        if hasattr(model, 'init_dataset'):
            for object in objects: model.init_dataset(object)
        setup_seed(option['seed']+346)
        # init communicator
        gv.communicator = flgo.VirtualCommunicator(objects)
        for ob in objects: ob.initialize()
    return objects[0]

def _init_with_meta(task_meta:dict, task: str, algorithm, option = {}, model=None, Logger: flgo.experiment.logger.BasicLogger = None, Simulator: BasicSimulator=flgo.simulator.DefaultSimulator, scene='horizontal'):
    r"""
    Initialize a runner in FLGo, which is to optimize a model on a specific task (i.e. IID-mnist-of-100-clients) by the selected federated algorithm.

    Args:
        task_meta (dict): the information to load task data from shared np.memmap
        task (str): the dictionary of the federated task
        algorithm (module|class): the algorithm will be used to optimize the model in federated manner, which must contain pre-defined attributions (e.g. algorithm.Server and algorithm.Client for horizontal federated learning)
        option (dict|str): the configurations of training, environment, algorithm, logger and simulator
        model (module|class): the model module that contains two methods: model.init_local_module(object) and model.init_global_module(object)
        Logger (flgo.experiment.logger.BasicLogger): the class of the logger inherited from flgo.experiment.logger.BasicLogger
        Simulator (flgo.simulator.base.BasicSimulator): the class of the simulator inherited from flgo.simulator.BasicSimulator
        scene (str): 'horizontal' or 'vertical' in current version of FLGo

    Returns:
        runner: the object instance that has the method object.run()

    Option:
        sample              (str):  methods for sampling client, which should be one of 'uniform' (default), 'md', 'full' or each with an additional suffix '_available', e.g., 'full_available'
        aggregate           (str):  methods for aggregating models,  which should be one of 'other' (default standard fedavg), 'uniform', 'weighted_scale', 'weighted_com'
        num_rounds          (int):  number of communication rounds, default is 20
        proportion          (flt):  proportion of clients sampled per round, default=0.2
        learning_rate_decay (flt):  learning rate decay for the training process, default=0.998
        lr_scheduler        (str):  type of the global learning rate scheduler, default='-1' (no decay), 0 is round-wise decay lr = lr*decay^round
        early_stop          (int):  stop training if there is no improvement for no smaller than the maximum rounds, default=-1 (no early stop)
        num_epochs          (int):  number of epochs when clients locally train the model on data, default=5
        num_step            (int):  the number of local steps, which dominate num_epochs when setting num_steps>0', default=-1
        learning_rate       (flt):  learning rate for inner solver;', type=float, default=0.1)
        batch_size          (int):  batch size', type=float, default='64')
        optimizer           (str):  select the optimizer for gd', type=str, choices=optimizer_list, default='SGD')
        clip_grad           (flt):  clipping gradients if the max norm of gradients ||g|| > clip_norm > 0, default=0.0
        momentum            (flt):  momentum of local training, default=0.0
        weight_decay        (flt):  weight decay of local training',default=0.0
        num_edge_rounds     (int):  number of edge rounds in hierFL',default=5
        algo_para           (any):  algorithm-dependent hyper-parameters
        train_holdout       (flt):  the rate of holding out the validation dataset from all the local training datasets',default=0.1
        test_holdout        (flt):  the rate of holding out the validation dataset from the testing datasets owned by the server', default=0.0
        local_test          (bool): if this term is set True and train_holdout>0, (0.5*train_holdout) of data will be set as client.test_data.default=False.
        seed                (flt):  seed for random initialization;', default=0
        dataseed            (int):  seed for random initialization for data train/val/test partition', default=0
        gpu                 (list\int): GPU IDs and empty input is equal to using CPU',
        server_with_cpu     (bool): the model parameters will be stored in the memory if True, default=False
        num_parallels       (int):  the number of parallels in the clients computing session, default=1
        parallel_type       (str):  the type of parallel: 't' means multi-threading and 'p' means multi-processing, default = 't'
        num_workers         (int):  the number of workers of DataLoader, default=0
        pin_memory          (bool): pin_memory of DataLoader, default=False
        drop_last           (bool): drop_last option of DataLoader, default is False
        test_batch_size     (int):  the batch_size used in testing phase, default=512
        availability        (str):  client availability mode, default = 'IDL'
        connectivity        (str):  client connectivity mode, default = 'IDL'
        completeness        (str):  client completeness mode, default = 'IDL'
        responsiveness      (str):  client responsiveness mode, default='IDL'
        log_level           (str):  the level of logger, default='INFO'
        log_file            (bool): whether log to file and default value is False, default=False
        no_log_console      (bool): whether log to screen and default value is True, default=False
        no_overwrite        (bool): whether to overwrite the old result, default=False
        eval_interval       (int):  evaluate every __ rounds, default=1
        save_checkpoint     (str):  the level of logger, default=''
        load_checkpoint     (str):  the level of logger, default=''

    Example:
    ```python
        >>> import flgo
        >>> from flgo.algorithm import fedavg
        >>> from flgo.experiment.logger.simple_logger import SimpleLogger
        >>> # create task 'mnist_iid' by flgo.gen_task('gen_config.yml', 'mnist_iid') if there exists no such task
        >>> if os.path.exists('mnist_iid'): flgo.gen_task({'benchmark':{'name':'flgo.benchmark.mnist_classification'}, 'partitioner':{'name':'IIDPartitioner','para':{'num_clients':100}}}, 'mnist_iid')
        >>> # create runner
        >>> fedavg_runner = flgo.init('mnist_iid', algorithm=fedavg, option = {'num_rounds':20, 'gpu':[0], 'learning_rate':0.1})
        >>> fedavg_runner.run()
        ... # the training will start after runner.run() was called, and the running-time results will be recorded by Logger into the task dictionary
    ```
    """
    # init option
    option = load_configuration(option)
    default_option = read_option_from_command()
    for op_key in option:
        if op_key in default_option.keys():
            op_type = type(default_option[op_key])
            if op_type == type(option[op_key]):
                default_option[op_key] = option[op_key]
            else:
                if op_type is list:
                    default_option[op_key]=list(option[op_key]) if hasattr(option[op_key], '__iter__') else [option[op_key]]
                elif op_type is tuple:
                    default_option[op_key] = tuple(option[op_key]) if hasattr(option[op_key], '__iter__') else (option[op_key])
                else:
                    default_option[op_key] = op_type(option[op_key])
        else:
            default_option[op_key] = option[op_key]
    option = default_option
    setup_seed(seed=option['seed'])
    option['task'] = task
    option['algorithm'] = (algorithm.__name__).split('.')[-1]
    option['scene'] = scene
    option['simulator'] = Simulator.__name__ if Simulator is not None else 'None'
    # # init devices
    # if not (option['gpu'] is None or len(option['gpu']) == 0) and option['gpu'][0]>0:
    #     os.environ['CUDA_VISIBLE_DEVICES'] = ','.join([str(i) for i in option['gpu']])
    #     option['gpu'] = list(range(len(option['gpu'])))
    # init task information
    if not os.path.exists(task):
        raise FileExistsError("Fedtask '{}' doesn't exist. Please generate the specified task by flgo.gen_task().")
    with open(os.path.join(task, 'info'), 'r') as inf:
        task_info = json.load(inf)
    # benchmark information
    benchmark = task_info['benchmark']
    if 'bmk_path' in task_info and os.path.isdir(task_info['bmk_path']): sys.path.append(task_info['bmk_path'])
    # model information
    if model== None:
        bmk_module = importlib.import_module(benchmark)
        if hasattr(algorithm, 'init_global_module') or hasattr(algorithm, 'init_local_module'):
            model = algorithm
        elif hasattr(bmk_module, 'default_model'):
            model = getattr(bmk_module, 'default_model')
        else:
            raise NotImplementedError("Model cannot be None when there exists no default model for the current benchmark {} and the algorithm {} didn't define the model by `init_local_module` or `init_global_module`".format(task_info['benchmark'], option['algorithm']))
    option['model'] = (model.__name__).split('.')[-1]

    # create global variable
    gv = GlobalVariable()
    # init logger
    default_scene_logger = {
        'horizontal': felp.SimpleLogger,
        'vertical': felp.VerticalLogger,
        'real_hclient': felp.ParallelHFLLogger,
        'real_hserver': felp.ParallelHFLLogger,
        'parallel_horizontal': felp.ParallelHFLLogger,
        'decentralized': felp.DecLogger,
        'hierarchical':felp.HierLogger,
    }
    assert scene in default_scene_logger.keys()
    if Logger is None: Logger = default_scene_logger[scene]
    logger = Logger(task=task, option=option, name=str(id(gv))+str(Logger), level=option['log_level'])
    gv.logger = logger

    # init device
    if (option['gpu'] is None or len(option['gpu']) == 0):
        gv.dev_list = [torch.device('cpu')]
    else:
        if int(option['gpu'][0])<0: # for mac equiped with M1
            gv.dev_list = [torch.device('mps')]
        else: # for other machines equiped with GPU
            gv.dev_list = [torch.device('cuda:{}'.format(gpu_id)) for gpu_id in option['gpu']]
    logger.info('PROCESS ID:\t{}'.format(os.getpid()))
    logger.info('Initializing devices: '+','.join([str(dev) for dev in gv.dev_list])+' will be used for this running.')

    # init task
    logger.info('BENCHMARK:\t{}'.format(benchmark))
    logger.info('TASK:\t\t\t{}'.format(task))
    logger.info('MODEL:\t\t{}'.format(model.__name__))
    logger.info('ALGORITHM:\t{}'.format(option['algorithm']))
    core_module = '.'.join([benchmark, 'core'])
    gv.TaskPipe = getattr(importlib.import_module(core_module), 'TaskPipe')
    task_pipe = gv.TaskPipe(task)
    TaskCalculator = getattr(importlib.import_module(core_module), 'TaskCalculator')
    gv.TaskCalculator = TaskCalculator
    setup_seed(option['dataseed'])

    # scene-specific procedure
    ############################################### Simulation FL ###########################################
    if scene in ['horizontal', 'vertical', 'decentralized', 'hierarchical','parallel_horizontal']:
        if option['load_mode']=='mmap':
            task_data = flgo.utils.shared_memory.load_task_data_from_npz(task, train_holdout=option['train_holdout'],
                                                                         test_holdout=option['test_holdout'],
                                                                         local_test=option['local_test'],
                                                                         local_test_ratio=option['local_test_ratio'],
                                                                         seed=option['dataseed'], create=True)
        else:
            task_data = fus.load_taskdata_from_memmap_meta(task_meta)
            # task_data = task_pipe.load_data(option)
            # if option['load_mode']=='mem':
            #     for u in task_data:
            #         for v in task_data[u]:
            #             task_data[u][v] = flgo.utils.shared_memory.MemDataset(task_data[u][v])
        # init objects
        obj_class = [c for c in dir(algorithm) if not c.startswith('__')]
        tmp = []
        for c in obj_class:
            try:
                C = getattr(algorithm, c)
                setattr(C, 'gv', gv)
                setattr(C, 'TaskCalculator', TaskCalculator)
                tmp.append(c)
            except:
                continue
        if scene == 'horizontal':
            for c in obj_class:
                if 'Client' in c:
                    class_client = getattr(algorithm, c)
                    class_client.train = flgo.simulator.base.with_completeness(class_client.train)
                elif 'Server' in c:
                    class_server = getattr(algorithm, c)
                    class_server.sample = flgo.simulator.base.with_availability(class_server.sample)
                    class_server.communicate_with = flgo.simulator.base.with_latency(class_server.communicate_with)
                    class_server.communicate = flgo.simulator.base.with_dropout(class_server.communicate)
        objects = task_pipe.generate_objects(option, algorithm, scene=scene)
        obj_classes = collections.defaultdict(int)
        for obj in objects: obj_classes[obj.__class__] += 1
        creating_str = []
        for k, v in obj_classes.items(): creating_str.append("{} {}".format(v, k))
        creating_str = ', '.join(creating_str)
        logger.info('SCENE:\t\t{} FL with '.format(scene) + creating_str)
        task_pipe.distribute(task_data, objects)

        # init model
        if hasattr(model, 'init_local_module'):
            for object in objects:
                model.init_local_module(object)
        if hasattr(model, 'init_global_module'):
            for object in objects:
                model.init_global_module(object)
        if hasattr(model, 'init_dataset'):
            for object in objects:
                model.init_dataset(object)
        setup_seed(option['seed'] + 346)

        # init communicator
        gv.communicator = flgo.VirtualCommunicator(objects)
        logger.info('SIMULATOR:\t{}'.format(str(Simulator)))
        gv.clock = flgo.simulator.base.ElemClock()
        gv.simulator = Simulator(objects, option) if scene == 'horizontal' else None
        if gv.simulator is not None: gv.simulator.initialize()
        gv.clock.register_simulator(simulator=gv.simulator)

        # final initialize
        for ob in objects: ob.initialize()

        # init virtual system environment
        logger.register_variable(coordinator=objects[0], participants=objects[1:], option=option, clock=gv.clock, scene=scene, objects=objects, simulator=Simulator.__name__ if scene == 'horizontal' else 'None')
        if scene == 'horizontal': logger.register_variable(server=objects[0], clients=objects[1:])
        logger.initialize()
        logger.info('Ready to start.')
        # register global variables for objects
        for c in tmp:
            try:
                C = getattr(algorithm, c)
                delattr(C, 'gv')
            except:
                continue
        for ob in objects:
            ob.gv = gv
        if gv.simulator is not None:
            gv.simulator.gv = gv
        gv.clock.gv = gv
        logger.gv = gv
        if scene=='parallel_horizontal':
            for obj in objects: obj.logger = Logger
            class HParallelRunner:
                def __init__(self, objects:list):
                    self.objects = objects
                    self.plist = []

                def run(self):
                    for obj in self.objects:
                        p = torch.multiprocessing.Process(target=obj.run, )
                        p.start()
                        self.plist.append(p)
            objects = [HParallelRunner(objects)]
    ############################################### Real World FL #####################################################
    else:
        raise RuntimeError(f"init_with_meta cannot support scecne {scene}. The supported scenes are {['horizontal', 'vertical', 'decentralized', 'hierarchical','parallel_horizontal']}")
    return objects[0]

def _call_by_process_with_meta(task_meta, task, algorithm_name,  opt, model_name, Logger, Simulator, scene, send_end):
    r"""
       This function is used to create a seperate child process.

       Args:
           task_meta (dict): the information to load task data from shared np.memmap
           task (str): the path of the task
           algorithm_name (str): the module name of algorithm
           opt (dict): option
           model_name (str): the module name of model
           Logger (flgo.experiment.logger.BasicLogger): the class of the logger
           Simulator (flgo.simulator.base.BasicSimulator): the class of the simulator inherited from flgo.simulator.BasicSimulator
           scene (str): horizontal or vertical
           send_end (connection.Connection): the return of multiprocess.Pipe(...) that is used to pass data to the parent process
       """

    pid = os.getpid()
    sys.stdout = open(os.devnull, 'w')
    if model_name is None:
        model = None
    else:
        try:
            model = importlib.import_module(model_name)
        except:
            model = model_name
    try:
        algorithm = importlib.import_module(algorithm_name)
    except:
        algorithm = algorithm_name
    es_key = 'val_loss'
    es_drct = -1
    try:
        runner = _init_with_meta(task_meta, task, algorithm, model=model, option=opt, Logger=Logger, Simulator=Simulator, scene=scene)
        es_key = runner.gv.logger.get_es_key()
        es_drct = runner.gv.logger.get_es_direction()
        runner.run()
        res = (
        os.path.join(runner.gv.logger.get_output_path(), runner.gv.logger.get_output_name()), es_key, es_drct, pid)
        send_end.send(res)
    except Exception as e:
        s = 'Process {} exits with error:" {}". '.format(pid, str(e))
        res = (opt, s, pid)
        send_end.send(res)

def _call_by_process(task, algorithm_name,  opt, model_name, Logger, Simulator, scene, send_end):
    r"""
    This function is used to create a seperate child process.

    Args:
        task (str): the path of the task
        algorithm_name (str): the module name of algorithm
        opt (dict): option
        model_name (str): the module name of model
        Logger (flgo.experiment.logger.BasicLogger): the class of the logger
        Simulator (flgo.simulator.base.BasicSimulator): the class of the simulator inherited from flgo.simulator.BasicSimulator
        scene (str): horizontal or vertical
        send_end (connection.Connection): the return of multiprocess.Pipe(...) that is used to pass data to the parent process
    """

    pid = os.getpid()
    sys.stdout = open(os.devnull, 'w')
    if model_name is None: model = None
    else:
        try:
            model = importlib.import_module(model_name)
        except:
            model = model_name
    try:
        algorithm = importlib.import_module(algorithm_name)
    except:
        algorithm = algorithm_name
    es_key = 'val_loss'
    es_drct = -1
    try:
        runner = flgo.init(task, algorithm, model=model, option=opt, Logger=Logger, Simulator=Simulator, scene=scene)
        es_key = runner.gv.logger.get_es_key()
        es_drct = runner.gv.logger.get_es_direction()
        runner.run()
        res = (os.path.join(runner.gv.logger.get_output_path(), runner.gv.logger.get_output_name()), es_key, es_drct, pid)
        send_end.send(res)
    except Exception as e:
        s = 'Process {} exits with error:" {}". '.format(pid, str(e))
        res = (opt, s, pid)
        send_end.send(res)

def run_in_parallel(task: str, algorithm, options:list = [], model=None, devices = [], Logger:flgo.experiment.logger.BasicLogger = flgo.experiment.logger.simple_logger.SimpleLogger, Simulator=flgo.simulator.DefaultSimulator, scene='horizontal', scheduler = None, mmap=False):
    """
    Run different groups of hyper-parameters for one task and one algorithm in parallel.

    Args:
        task (str): the dictionary of the federated task
        algorithm (module|class): the algorithm will be used to optimize the model in federated manner, which must contain pre-defined attributions (e.g. algorithm.Server and algorithm.Client for horizontal federated learning)
        options (list): the configurations of different groups of hyper-parameters
        model (module|class): the model module that contains two methods: model.init_local_module(object) and model.init_global_module(object)
        devices (list): the list of IDs of devices
        Logger (class): the class of the logger inherited from flgo.experiment.logger.BasicLogger
        Simulator (class): the class of the simulator inherited from flgo.simulator.BasicSimulator
        scene (str): 'horizontal' or 'vertical' in current version of FLGo
        scheduler (instance of flgo.experiment.device_scheduler.BasicScheduler): GPU scheduler that schedules GPU by checking their availability
        mmap (bool): whether to load all the dataset into the shared memory where all the processes can directly access the dataset

    Returns:
        the returns of _call_by_process
    """
    try:
        # init multiprocess
        torch.multiprocessing.set_start_method('spawn', force=True)
        torch.multiprocessing.set_sharing_strategy('file_system')
    except:
        pass
    if model is None:
        model_name = None
    else:
        if not hasattr(model, '__module__') and hasattr(model, '__name__'):
            model_name = model.__name__
        else:
            model_name = model
    algorithm_name = algorithm.__name__ if (not hasattr(algorithm, '__module__') and hasattr(algorithm, '__name__')) else algorithm
    option_state = {oid:{'p':None, 'completed':False, 'output':None, 'option_in_queue':False, 'recv':None, } for oid in range(len(options))}
    if scheduler is None: scheduler = flgo.experiment.device_scheduler.BasicScheduler(devices)
    es_key = None
    es_drct = None
    task_meta = None
    if mmap:
        load_option = options[0] if len(options)>0 else {}
        keywords = ['train_holdout', 'test_holdout', 'local_test', 'local_test_ratio']
        task_data = load_task_data(task, **{k:v for k,v in load_option.items() if k in keywords})
        cache_path = os.path.join(task, '.cache')
        if not os.path.exists(cache_path): os.mkdir(cache_path)
        task_meta = fus.create_memmap_meta_for_task(task_data, os.path.abspath(cache_path))
    while True:
        for oid in range(len(options)):
            opt = options[oid]
            if option_state[oid]['p'] is None:
                if not option_state[oid]['completed']:
                    available_device = scheduler.get_available_device(opt)
                    if available_device is None: continue
                    else:
                        opt['gpu'] = available_device
                        recv_end, send_end = multiprocessing.Pipe(False)
                        if mmap:
                            option_state[oid]['p'] = multiprocessing.Process(target=_call_by_process_with_meta, args=(task_meta, task, algorithm_name, opt, model_name, Logger, Simulator, scene, send_end))
                        else:
                            option_state[oid]['p'] = multiprocessing.Process(target=_call_by_process, args=(task, algorithm_name, opt, model_name, Logger, Simulator, scene, send_end))
                        option_state[oid]['recv'] = recv_end
                        option_state[oid]['p'].start()
                        scheduler.add_process(option_state[oid]['p'].pid)
                        print('Process {} was created for args {}'.format(option_state[oid]['p'].pid,(task, algorithm_name, opt, model_name, Logger, Simulator, scene)))
            else:
                if option_state[oid]['p'].exitcode is not None:
                    tmp = option_state[oid]['recv'].recv()
                    scheduler.remove_process(tmp[-1])
                    try:
                        option_state[oid]['p'].terminate()
                    except:
                        pass
                    option_state[oid]['p'] = None
                    if len(tmp)==4:
                        option_state[oid]['completed'] = True
                        option_state[oid]['output'] = tmp[0]
                        if es_key is None: es_key = tmp[1]
                        if es_drct is None: es_drct = tmp[2]
                    else:
                        print(tmp[1])
                        if "All the received local models have parameters of nan value." in tmp[1]:
                            option_state[oid]['completed'] = True
                            option_state[oid]['output'] = tmp[1]
        if all([v['completed'] for v in option_state.values()]):break
        time.sleep(1)
    res = []
    for oid in range(len(options)):
        rec_path = option_state[oid]['output']
        if os.path.exists(rec_path):
            with open(rec_path, 'r') as inf:
                s_inf = inf.read()
                rec = json.loads(s_inf)
            res.append(rec)
    return res, es_key, es_drct

def run_in_sequencial(task: str, algorithm, options:list = [], model=None, Logger:flgo.experiment.logger.BasicLogger = flgo.experiment.logger.simple_logger.SimpleLogger, Simulator=flgo.simulator.DefaultSimulator, scene='horizontal', mmap=False):
    """
    Run different groups of hyper-parameters for one task and one algorithm in parallel.

    Args:
        task (str): the dictionary of the federated task
        algorithm (module|class): the algorithm will be used to optimize the model in federated manner, which must contain pre-defined attributions (e.g. algorithm.Server and algorithm.Client for horizontal federated learning)
        options (list): the configurations of different groups of hyper-parameters
        model (module|class): the model module that contains two methods: model.init_local_module(object) and model.init_global_module(object)
        Logger (class): the class of the logger inherited from flgo.experiment.logger.BasicLogger
        Simulator (class): the class of the simulator inherited from flgo.simulator.BasicSimulator
        scene (str): 'horizontal' or 'vertical' in current version of FLGo
        mmap (bool): whether to load all the dataset into the shared memory where all the processes can directly access the dataset

    Returns:
        the running logs
    """
    es_key = None
    es_drct = None
    res = []
    load_mode = options[0].get('load_mode', None)
    if mmap and load_mode not in ['mem', 'mmap']:
        first_op = options[0]
        keywords = ['train_holdout', 'test_holdout', 'local_test', 'local_test_ratio', 'dataseed']
        first_op = {k: v for k, v in first_op.items() if k in keywords}
        if first_op.get('dataseed', None) is not None:
            first_op['seed'] = first_op['dataseed']
            first_op.pop('dataseed')
        task_data = load_task_data(task, **first_op)
        cache_path = os.path.join(task, '.cache')
        if not os.path.exists(cache_path): os.mkdir(cache_path)
        task_meta = fus.create_memmap_meta_for_task(task_data, os.path.abspath(cache_path))
        for op in options:
            runner = _init_with_meta(task_meta, task, algorithm, op, model, Logger=Logger, Simulator=Simulator, scene=scene)
            if es_key is None:
                es_key = runner.gv.logger.get_es_key()
                es_drct = runner.gv.logger.get_es_direction()
            runner.run()
            res.append(os.path.join(runner.gv.logger.get_output_path(), runner.gv.logger.get_output_name()))
            del runner
    else:
        for op in options:
            runner = flgo.init(task, algorithm, op, model, Logger=Logger, Simulator=Simulator, scene=scene)
            if es_key is None:
                es_key = runner.gv.logger.get_es_key()
                es_drct = runner.gv.logger.get_es_direction()
            runner.run()
            res.append(os.path.join(runner.gv.logger.get_output_path(), runner.gv.logger.get_output_name()))
            del runner
    outputs = []
    for oid in range(len(options)):
        rec_path = res[oid]
        if os.path.exists(rec_path):
            with open(rec_path, 'r') as inf:
                s_inf = inf.read()
                rec = json.loads(s_inf)
            outputs.append(rec)
    return outputs, es_key, es_drct


def tune(task: str, algorithm, option: dict = {}, model=None, Logger: flgo.experiment.logger.BasicLogger = flgo.experiment.logger.tune_logger.TuneLogger, Simulator: BasicSimulator=flgo.simulator.DefaultSimulator, scene='horizontal', scheduler=None, mmap=False):
    """
        Tune hyper-parameters for the specific (task, algorithm, model) in parallel.
        Args:
            task (str): the dictionary of the federated task
            algorithm (module|class): the algorithm will be used to optimize the model in federated manner, which must contain pre-defined attributions (e.g. algorithm.Server and algorithm.Client for horizontal federated learning)
            option (dict): the dict whose values should be of type list to construct the combinations
            model (module|class): the model module that contains two methods: model.init_local_module(object) and model.init_global_module(object)
            Logger (class): the class of the logger inherited from flgo.experiment.logger.BasicLogger
            Simulator (class): the class of the simulator inherited from flgo.simulator.BasicSimulator
            scene (str): 'horizontal' or 'vertical' in current version of FLGo
            scheduler (instance of flgo.experiment.device_scheduler.BasicScheduler): GPU scheduler that schedules GPU by checking their availability
            mmap (bool): whether to load all the dataset into the shared memory where all the processes can directly access the dataset
        """
    # generate combinations of hyper-parameters
    if 'gpu' in option.keys():
        device_ids = option['gpu']
        option.pop('gpu')
        if not isinstance(device_ids, Iterable): device_ids = [device_ids]
    else:
        device_ids = [-1]
    keys = list(option.keys())
    for k in keys: option[k] = [option[k]] if (not isinstance(option[k], Iterable) or isinstance(option[k], str)) else option[k]
    para_combs = [para_comb for para_comb in itertools.product(*(option[k] for k in keys))]
    options = [{k:v for k,v in zip(keys, paras)} for paras in para_combs]
    for op in options:op['log_file'] = True
    if scheduler is None:
        scheduler = flgo.experiment.device_scheduler.BasicScheduler(device_ids)
    outputs, es_key, es_drct = run_in_parallel(task, algorithm, options,model, devices=device_ids, Logger=Logger, Simulator=Simulator, scene=scene, scheduler=scheduler, mmap=mmap)
    if len(outputs)==0:
        warnings.warn("All the groups of parameters had resulted in divergence of model training")
        return {}
    if es_drct<=0:
        optimal_idx = int(np.argmin([min(output[es_key]) for output in outputs]))
    else:
        optimal_idx = int(np.argmax([max(output[es_key]) for output in outputs]))
    optimal_para = options[optimal_idx]
    print("The optimal combination of hyper-parameters is:")
    print('-----------------------------------------------')
    for k,v in optimal_para.items():
        if k=='gpu': continue
        print("{}\t|{}".format(k,v))
    print('-----------------------------------------------')
    op_round = np.argmin(outputs[optimal_idx][es_key]) if es_drct<=0 else np.argmax(outputs[optimal_idx][es_key])
    op_value = np.min(outputs[optimal_idx][es_key]) if es_drct<=0 else np.max(outputs[optimal_idx][es_key])
    if 'eval_interval' in option.keys(): op_round = option['eval_interval']*op_round
    print('The optimal value {} of {} occurs at the round {}'.format(op_value, es_key, op_round))
    return optimal_para

def tune_sequencially(task: str, algorithm, option: dict = {}, model=None, Logger: flgo.experiment.logger.BasicLogger = flgo.experiment.logger.tune_logger.TuneLogger, Simulator: BasicSimulator=flgo.simulator.DefaultSimulator, scene='horizontal', mmap=False):
    """
        Tune hyper-parameters for the specific (task, algorithm, model) in sequencial.
        Args:
            task (str): the dictionary of the federated task
            algorithm (module|class): the algorithm will be used to optimize the model in federated manner, which must contain pre-defined attributions (e.g. algorithm.Server and algorithm.Client for horizontal federated learning)
            option (dict): the dict whose values should be of type list to construct the combinations
            model (module|class): the model module that contains two methods: model.init_local_module(object) and model.init_global_module(object)
            Logger (class): the class of the logger inherited from flgo.experiment.logger.BasicLogger
            Simulator (class): the class of the simulator inherited from flgo.simulator.BasicSimulator
            scene (str): 'horizontal' or 'vertical' in current version of FLGo
            mmap (bool): whether to load all the dataset into the shared memory where all the processes can directly access the dataset
        """
    # generate combinations of hyper-parameters
    keys = list(option.keys())
    for k in keys: option[k] = [option[k]] if (not isinstance(option[k], Iterable) or isinstance(option[k], str)) else option[k]
    para_combs = [para_comb for para_comb in itertools.product(*(option[k] for k in keys))]
    options = [{k:v for k,v in zip(keys, paras)} for paras in para_combs]
    for op in options:op['log_file'] = True
    es_key, es_drct = None, None
    res = []
    load_mode = option.get('load_mode', None)
    if mmap and load_mode not in ['mem', 'mmap']:
        first_op = options[0]
        keywords = ['train_holdout', 'test_holdout', 'local_test', 'local_test_ratio', 'dataseed']
        first_op = {k: v for k, v in first_op.items() if k in keywords}
        if first_op.get('dataseed', None) is not None:
            first_op['seed'] = first_op['dataseed']
            first_op.pop('dataseed')
        task_data = load_task_data(task, **first_op)
        cache_path = os.path.join(task, '.cache')
        if not os.path.exists(cache_path): os.mkdir(cache_path)
        task_meta = fus.create_memmap_meta_for_task(task_data, os.path.abspath(cache_path))
        for op in options:
            runner = _init_with_meta(task_meta, task, algorithm, op, model, Logger=Logger, Simulator=Simulator, scene=scene)
            if es_key is None:
                es_key = runner.gv.logger.get_es_key()
                es_drct = runner.gv.logger.get_es_direction()
            runner.run()
            res.append(os.path.join(runner.gv.logger.get_output_path(), runner.gv.logger.get_output_name()))
            del runner
    else:
        for op in options:
            runner = flgo.init(task, algorithm, op, model, Logger=Logger, Simulator=Simulator, scene=scene)
            if es_key is None:
                es_key = runner.gv.logger.get_es_key()
                es_drct = runner.gv.logger.get_es_direction()
            runner.run()
            res.append(os.path.join(runner.gv.logger.get_output_path(), runner.gv.logger.get_output_name()))
            del runner
    outputs = []
    for oid in range(len(options)):
        rec_path = res[oid]
        if os.path.exists(rec_path):
            with open(rec_path, 'r') as inf:
                s_inf = inf.read()
                rec = json.loads(s_inf)
            outputs.append(rec)
    if len(outputs)==0:
        warnings.warn("All the groups of parameters had resulted in divergence of model training")
        return {}
    if es_drct<=0:
        optimal_idx = int(np.argmin([min(output[es_key]) for output in outputs]))
    else:
        optimal_idx = int(np.argmax([max(output[es_key]) for output in outputs]))
    optimal_para = options[optimal_idx]
    print("The optimal combination of hyper-parameters is:")
    print('-----------------------------------------------')
    for k,v in optimal_para.items():
        if k=='gpu': continue
        print("{}\t|{}".format(k,v))
    print('-----------------------------------------------')
    op_round = np.argmin(outputs[optimal_idx][es_key]) if es_drct<=0 else np.argmax(outputs[optimal_idx][es_key])
    op_value = np.min(outputs[optimal_idx][es_key]) if es_drct<=0 else np.max(outputs[optimal_idx][es_key])
    if 'eval_interval' in option.keys(): op_round = option['eval_interval']*op_round
    print('The optimal value {} of {} occurs at the round {}'.format(op_value, es_key, op_round))
    return optimal_para


def multi_init_and_run(runner_args:list, devices = [], scheduler=None, mmap=False):
    r"""
    Create multiple runners and run in parallel

    Args:
        runner_args (list): each element in runner_args should be either a dict or a tuple or parameters
        devices (list): a list of gpu id
        scheduler (flgo.experiment.device_scheduler.BasicScheduler(...)): GPU scheduler
        mmap (bool): whether to load all the dataset into the shared memory where all the processes can directly access the dataset

    Returns:
        a list of output results of runners

    Example:
    ```python
        >>> from flgo.algorithm import fedavg, fedprox, scaffold
        >>> # create task 'mnist_iid' by flgo.gen_task if there exists no such task
        >>> task='./mnist_iid'
        >>> if os.path.exists(task): flgo.gen_task({'benchmark':{'name':'flgo.benchmark.mnist_classification'}, 'partitioner':{'name':'IIDPartitioner','para':{'num_clients':100}}}, task)
        >>> algos = [fedavg, fedprox, scaffold]
        >>> flgo.multi_init_and_run([{'task':task, 'algorithm':algo} for algo in algos], devices=[0])
    ```
    """
    if len(runner_args)==0:return
    args = []
    if type(runner_args[0]) is dict:
        keys = ['task', 'algorithm', 'option', 'model', 'Logger', 'Simulator', 'scene']
        for a in runner_args:
            tmp = collections.defaultdict(lambda:None, a)
            if tmp['task'] is None or tmp['algorithm'] is None:
                raise RuntimeError("keyword 'task' or 'algorithm' is of NoneType")
            algorithm = tmp['algorithm']
            tmp['algorithm'] = algorithm.__name__ if (not hasattr(algorithm, '__module__') and hasattr(algorithm, '__name__')) else algorithm
            if tmp['option'] is None:
                tmp['option'] = default_option_dict
            else:
                option = tmp['option']
                default_option = read_option_from_command()
                for op_key in option:
                    if op_key in default_option.keys():
                        op_type = type(default_option[op_key])
                        if op_type == type(option[op_key]):
                            default_option[op_key] = option[op_key]
                        else:
                            if op_type is list:
                                default_option[op_key] = list(option[op_key]) if hasattr(option[op_key],
                                                                                         '__iter__') else [
                                    option[op_key]]
                            elif op_type is tuple:
                                default_option[op_key] = tuple(option[op_key]) if hasattr(option[op_key],
                                                                                          '__iter__') else (
                                option[op_key])
                            else:
                                default_option[op_key] = op_type(option[op_key])
                tmp['option'] = default_option
            if tmp['model'] is None:
                model_name = None
            else:
                if not hasattr(tmp['model'], '__module__') and hasattr(tmp['model'], '__name__'):
                    model_name = tmp['model'].__name__
                else:
                    model_name = tmp['model']
            tmp['model'] = model_name
            if tmp['Logger'] is None:
                tmp['Logger'] = flgo.experiment.logger.simple_logger.SimpleLogger
            algorithm_name = tmp['algorithm'].__name__ if (not hasattr(tmp['algorithm'], '__module__') and hasattr(tmp['algorithm'], '__name__')) else tmp['algorithm']
            if tmp['Simulator'] is None:
                tmp['Simulator'] = flgo.simulator.DefaultSimulator
            if tmp['scene'] is None:
                tmp['scene'] = 'horizontal'
            args.append([tmp[k] for k in keys])
    elif type(runner_args[0]) is tuple or type(runner_args[0]) is list:
        for a in runner_args:
            if len(a)<2: raise RuntimeError('the args of runner should at least contain task and algorithm.')
            default_args = [None, None, default_option_dict, None, flgo.experiment.logger.simple_logger.SimpleLogger, flgo.simulator.DefaultSimulator, 'horizontal']
            for aid in range(len(a)):
                if aid==0:
                    default_args[aid] = a[aid]
                if aid==1:
                    algorithm = a[aid]
                    algorithm_name = algorithm.__name__ if (not hasattr(algorithm, '__module__') and hasattr(algorithm, '__name__')) else algorithm
                    default_args[aid] = algorithm_name
                elif aid==2:
                    option = a[aid]
                    default_option = read_option_from_command()
                    for op_key in option:
                        if op_key in default_option.keys():
                            op_type = type(default_option[op_key])
                            if op_type == type(option[op_key]):
                                default_option[op_key] = option[op_key]
                            else:
                                if op_type is list:
                                    default_option[op_key] = list(option[op_key]) if hasattr(option[op_key],
                                                                                             '__iter__') else [
                                        option[op_key]]
                                elif op_type is tuple:
                                    default_option[op_key] = tuple(option[op_key]) if hasattr(option[op_key],
                                                                                              '__iter__') else (
                                        option[op_key])
                                else:
                                    default_option[op_key] = op_type(option[op_key])
                    default_args[aid] = default_option
                elif aid==3:
                    model = a[aid]
                    if model is None:
                        model_name = None
                    else:
                        if not hasattr(model, '__module__') and hasattr(model, '__name__'):
                            model_name = model.__name__
                        else:
                            model_name = model
                    default_args[aid] = model_name
                else:
                    default_args[aid] = a[aid]

    runner_state = {rid: {'p': None, 'completed': False, 'output': None, 'runner_in_queue': False, 'recv': None, } for
                    rid in range(len(args))}
    if scheduler is None: scheduler = flgo.experiment.device_scheduler.BasicScheduler(devices)
    task_meta_dict = {}
    if mmap:
        all_tasks = list(set([rarg['task'] for rarg in runner_args]))
        keywords = ['train_holdout', 'test_holdout', 'local_test', 'local_test_ratio']
        task_meta_dict = {}
        for task in all_tasks:
            load_option = {}
            for rarg in runner_args:
                if rarg['task']==task:
                    load_option = rarg['option']
                    break
            task_cache = os.path.join(task, '.cache')
            if not os.path.exists(task_cache): os.mkdir(task_cache)
            task_meta_dict[task] = fus.create_memmap_meta_for_task(load_task_data(task, **{k:v for k,v in load_option.items() if k in keywords}), task_cache)
    while True:
        for rid in range(len(args)):
            current_arg = args[rid]
            if runner_state[rid]['p'] is None:
                if not runner_state[rid]['completed']:
                    available_device = scheduler.get_available_device(current_arg)
                    if available_device is None:
                        continue
                    else:
                        list_current_arg = copy.deepcopy(current_arg)
                        list_current_arg[2]['gpu'] = available_device
                        recv_end, send_end = multiprocessing.Pipe(False)
                        list_current_arg.append(send_end)
                        if mmap:
                            list_current_arg = [task_meta_dict[current_arg[0]]] + list_current_arg
                            _call_func = _call_by_process_with_meta
                        else:
                            _call_func = _call_by_process
                        runner_state[rid]['p'] = multiprocessing.Process(target=_call_func, args=tuple(list_current_arg))
                        runner_state[rid]['recv'] = recv_end
                        runner_state[rid]['p'].start()
                        scheduler.add_process(runner_state[rid]['p'].pid)
                        print('Process {} was created for args {}'.format(runner_state[rid]['p'].pid,current_arg))
            else:
                if runner_state[rid]['p'].exitcode is not None:
                    tmp = runner_state[rid]['recv'].recv()
                    scheduler.remove_process(tmp[-1])
                    try:
                        runner_state[rid]['p'].terminate()
                    except:
                        pass
                    runner_state[rid]['p'] = None
                    if len(tmp) == 4:
                        runner_state[rid]['completed'] = True
                        runner_state[rid]['output'] = tmp[0]
                    else:
                        print(tmp[1])
                        if "All the received local models have parameters of nan value." in tmp[1]:
                            runner_state[rid]['completed'] = True
                            runner_state[rid]['output'] = tmp[1]
        if all([v['completed'] for v in runner_state.values()]): break
        time.sleep(1)
    res = []
    for rid in range(len(runner_state)):
        rec_path = runner_state[rid]['output']
        with open(rec_path, 'r') as inf:
            s_inf = inf.read()
            rec = json.loads(s_inf)
        res.append(rec)
    return res

def convert_model(get_model:Callable, model_name='anonymous', scene:str='horizontal', attr_preversed=True):
    r"""
    Convert an existing model into a model that can be loaded in flgo.
    Args:
        get_model (Callable): this function will return a model of type torch.nn.Module when it is called
        model_name (str): the name of the model
        scene (str): the FL scene

    Returns:
        res_model: the model can be used in flgo.init(..., model=res_model, ...)
    """
    if attr_preversed and not isinstance(get_model, types.FunctionType):
        class DecoratedModel(get_model, flgo.utils.fmodule.FModule):
            def __init__(self, *args, **kwargs):
                flgo.utils.fmodule.FModule.__init__(self)
                super().__init__(*args, **kwargs)
    else:
        class DecoratedModel(flgo.utils.fmodule.FModule):
            def __init__(self):
                super().__init__()
                self.model = get_model()

            def forward(self, *args, **kwargs):
                return self.model(*args, **kwargs)

    if scene=='horizontal':
        class AnonymousModel:
            __name__ = model_name

            @classmethod
            def init_global_module(self, object):
                if 'Server' in object.__class__.__name__:
                    object.model = DecoratedModel().to(object.device)

            @classmethod
            def init_local_module(self, object):
                pass

    elif scene=='decentralized':
        class AnonymousModel:
            __name__ = model_name

            @classmethod
            def init_local_module(self, object):
                if 'Client' in object.__class__.__name__:
                    object.model = DecoratedModel().to(object.device)

            @classmethod
            def init_global_module(self, object):
                pass
    else:
        raise NotImplementedError('The current version only support converting model for horizontalFL and DecentralizedFL.')
    return AnonymousModel()

def module2fmodule(Model):
    """
    Convert a class of torch.nn.Module into class flgo.utils.fmodule.FModule
    Args:
        Model (class): a class inherited from torch.nn.Module

    Returns:
        TempModule (class): The same class but additionally inheriting from flgo.utils.fmodule.FModule

    """
    class TempFModule(Model, flgo.utils.fmodule.FModule):
        def __init__(self, *args, **kwargs):
            super(TempFModule, self).__init__(*args, **kwargs)
    return TempFModule

def set_data_root(data_root:str=None):
    """
    Set the root of data that stores all the raw data automatically
    Args:
        data_root (str): the path of a directory whose default value is None and will be set to os.getcwd() as default.
    Returns:
    """
    file_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'benchmark', '__init__.py')
    default_root = os.path.abspath(os.path.join(flgo.benchmark.path, 'RAW_DATA'))
    if data_root is None and os.path.abspath(flgo.benchmark.data_root)!=default_root:
        crt_root = default_root.strip()
        root_name = '"'+default_root.strip()+'"'
    elif data_root == 'cwd':
        crt_root = os.path.abspath(os.getcwd())
        root_name = 'os.getcwd()'
    else:
        if not os.path.exists(data_root):
            os.makedirs(data_root)
        if not os.path.isdir(data_root):
            raise TypeError('data_root must be a dir')
        crt_root = os.path.abspath(data_root).strip()
        root_name = '"'+crt_root+'"'
    with open(file_path, 'r', encoding=sys.getfilesystemencoding()) as inf:
        lines = inf.readlines()
        idx = -1
        for i,line in enumerate(lines):
            if line.find('data_root')>-1:
                idx = i
                break
        if idx>0:
            lines[idx] = "data_root = "+ root_name
    with open(file_path, 'w', encoding=sys.getfilesystemencoding()) as outf:
        outf.writelines(lines)
    flgo.benchmark.data_root = crt_root
    print('Data root directory has successfully been changed to {}'.format(crt_root))
    return

def download_resource(root:str, name:str, type:str, overwrite:bool=False):
    """
    Download resource from github
    Args:
        root (str): the path to store the resource
        name (str): the name of the resource
        type (type): the type of the resource in ['algorithm', 'benchmark', 'simulator']
        overwrite (bool): whether to overwrite existing file
    Returns:
    """
    resource_root = "https://github.com/WwZzz/easyFL/raw/FLGo/resources/"
    if type not in ['algorithm', 'benchmark', 'simulator']: raise ValueError("Args type must of value in ['algorithm', 'benchmark', 'simulator']")
    url = resource_root+type+'/'
    suffix_dict = {'algorithm':'.py', 'simulator':'.py', 'benchmark':'.zip'}
    suffix = suffix_dict[type]
    file_name = name
    if not file_name.endswith(suffix):
        file_name = file_name+suffix
    if not os.path.exists(file_name) or overwrite:
        try:
            urllib.request.urlretrieve(url+file_name, os.path.join(root, file_name))
        except Exception as e:
            print(e)
            return None
    else:
        warnings.warn("There already exist {} named {}".format(type, name))
    if type == 'benchmark':
        bmk_zip = zipfile.ZipFile(file_name)
        bmk_zip.extractall(root)
    module_path = '.'.join(os.path.relpath(os.path.join(root, name), os.path.curdir).split(os.path.sep))
    module = importlib.import_module(module_path)
    if type in ['algorithm', 'benchmark']:
        return module
    else:
        Simulator = getattr(module, 'Simulator')
        return Simulator

def list_resource(type:str='algorithm'):
    """
    List currently available resources at github. The arg. `type` should be one of elements in {'algorithm', 'benchmark', 'simulator'}
    Args:
        type (str):
    Returns:
        res (list): the name of currently available resources
    """
    if type not in ['algorithm', 'benchmark', 'simulator']: raise ValueError("Args type must of value in ['algorithm', 'benchmark', 'simulator']")
    url = "https://github.com/WwZzz/easyFL/tree/FLGo/resources/"+type
    suffix_dict = {"algorithm": ".py", "benchmark":".zip", "simulator":".py"}
    suffix = suffix_dict[type]
    try:
        content = str(requests.get(url).content, encoding=sys.getfilesystemencoding())
    except Exception as e:
        print(e)
        return None
    res = re.findall(r'"[a-zA-Z0-9_-]*{}"'.format(suffix), content)
    res = [s.strip('"') for s in res]
    res = [s[:-len(suffix)] for s in res if s!="test.py"]
    return res

def gen_empty_task(benchmark, task_path:str, scene:str="unknown"):
    r"""
    Create empty task for a benchmark without information about partitioning.
    Args:
        benchmark (module): the benchmark module
        task_path (str): the path of the task
        scene (str): the scene of FL
    Return:
        task_path
    """
    if os.path.exists(task_path):
        warnings.warn("Task {} already exists.".format(task_path))
        return task_path
    info = {'benchmark':benchmark.__name__, "scene":scene, 'bmk_path': os.path.dirname(benchmark.__file__)}
    os.makedirs(os.path.join(task_path, 'log'))
    os.makedirs(os.path.join(task_path, 'record'))
    with open(os.path.join(task_path, 'info'), 'w') as outinfo:
        json.dump(info, outinfo)
    with open(os.path.join(task_path, 'dataset.py'), 'w') as outf:
        pass
    print("Empty task {} has been successfully generated.".format(task_path))
    return task_path

def zip_task(task_path:str, target_path='.', with_bmk:bool=True, ignore_names:list=['data.json', '__pycache__', 'checkpoint']):
    """
        Compress an existing task folder into a .zip file. The zipped task can be transmitted to others.
    Args:
        task_path (str): the task path
        target_path= (str): the target directory to save the zipped file
        with_bmk (bool): whether to zip the benchmark codes together
    Return:
        output_path (str): the path of the zipped task
    """
    if not os.path.exists(task_path):
        raise FileNotFoundError("Task {} doesn't exist.".format(task_path))
    if not os.path.isdir(target_path):
        raise TypeError("target path must be a directory.")
    if not os.path.isdir(task_path):
        raise TypeError("task path must be a directory.")
    with open(os.path.join(task_path, 'info'), 'r') as inf:
        info = json.load(inf)
    old_info = info.copy()
    config_path = os.path.join(task_path, 'dataset.py')
    # if os.path.exists(config_path):
    #     with open(config_path, 'r') as in_dataset:
    #         old_config = in_dataset.readlines()
    # with open(config_path, 'w') as new_dataset:
    #     new_dataset.writelines(['train_data=None\n', 'test_data=None\n', 'val_data=None\n'])
    output_path = os.path.join(target_path, os.path.basename(task_path)+'.zip')
    zipf = zipfile.ZipFile(output_path, 'w', zipfile.ZIP_DEFLATED)

    flag = False
    if with_bmk:
        bmk = importlib.import_module(info['benchmark'])
        bmk_dir = os.path.dirname(bmk.__file__)
        bmk_base = os.path.basename(bmk_dir)
        if 'bmk_path' in info:
            sys.path.append(info['bmk_path'])
            info['bmk_path'] = '.'
            info['benchmark'] = bmk_base
            with open(os.path.join(task_path, 'info'), 'w') as outf:
                json.dump(info, outf)
            flag = True
        for root, dirs, files in os.walk(bmk_dir):
            relative_root = bmk_base if root == bmk_dir else root.replace(bmk_dir, bmk_base) + os.sep
            for filename in files:
                zipf.write(os.path.join(root, filename), os.path.join(relative_root,filename))
    task_base = os.path.basename(task_path)
    for root, dirs, files in os.walk(task_path):
        relative_root = task_base if root == task_path else root.replace(task_path, task_base) + os.sep
        for filename in files:
            p = os.path.join(relative_root,filename)
            if any([k in p for k in ignore_names]): continue
            zipf.write(os.path.join(root, filename), p)
    zipf.close()
    if flag:
        with open(os.path.join(task_path, 'info'), 'w') as outf:
            json.dump(old_info, outf)
    # with open(config_path, 'w') as out_dataset:
    #     out_dataset.writelines(old_config)
    return output_path

def pull_task_from_(address:str, task_name:str, target_path='.', unzip=True, timeout=10):
    """
    Pull task from the server at given ip address. The pulled task will be a zip file that can be extracted to the federated task in flgo.
    Args:
        address (str): the ip address of the server
        task_name= (str): the name of the task
        target_path= (str): the target directory to save the zipped file
        unzip (bool): whether to unzip the pulled task
        timeout (int): connections of timeout
    Return:
        output_path (str): the path of the zipped task
    """
    if os.path.exists(os.path.join(target_path, task_name)):
        warnings.warn("Task {} already exists. Please remove this task before pulling.".format(os.path.join(target_path, task_name)))
        return
    task_zip = os.path.basename(task_name)+'.zip'
    zip_path = os.path.join(target_path, task_zip)
    if os.path.exists(zip_path):
        raise FileExistsError("Zipped task {} already exists.".format(zip_path))
    # ctx = zmq.Context()
    sck = _ctx.socket(zmq.REQ)
    sck.setsockopt(zmq.RCVTIMEO, timeout*1000)
    sck.connect(address)
    try:
        print('Requesting task %s from %s ...'% (task_name, address))
        sck.send_string(_get_name(), zmq.SNDMORE)
        sck.send(b"pull task")
        chunk = sck.recv()
        print('Successfully received task from %s'%address)
    except Exception as e:
        print(e)
        return
    with open(zip_path, 'wb') as f:
        f.write(chunk)
    assert os.path.exists(zip_path)
    # unzip
    if unzip:
        zip_task = zipfile.ZipFile(zip_path, 'r')
        zip_task.extractall(target_path)
        os.remove(zip_path)
    return

def _get_name():
    import uuid
    import socket
    if flgo._name is None:
        try:
            ip = requests.get('https://api.ipify.org').text
        except:
            ip = 'unknown_ip'
        flgo._name = '_'.join([ip, socket.gethostname(), str(uuid.getnode())])
    return flgo._name

def clear_task_cache(task):
    if not os.path.exists(task): return
    cache = os.path.join(task, '.cache')
    if os.path.exists(cache):
        shutil.rmtree(cache)


from ..config import get_model
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self):
        super().__init__()
        self.model = get_model()

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import torch.nn as nn
from flgo.utils.fmodule import FModule
import torch
import numpy as np

class RecEmbedding(FModule):
    def __init__(self, dim=1682, embedding_size=20):
        super().__init__()
        self.embeddings = nn.Embedding(dim, embedding_size)

    def get_embedding(self, item_index=None):
        if item_index is None: return self.embeddings.weight
        return self.embeddings.weight[item_index]

    def set_embedding(self, array):
        assert tuple(self.embeddings.weight.shape)==array.shape
        with torch.no_grad():
            self.embeddings.weight.data = torch.tensor(array, dtype=torch.float32)

# globally shared item vectors
class ItemEmbedding(RecEmbedding):
    def __init__(self, n_items=1682, embedding_size=20):
        super().__init__(dim=n_items, embedding_size=embedding_size)

    def forward(self, batch_data, local_model):
        user_id = batch_data['user_id'].long()
        item_id = batch_data['item_id'].long()
        user_id_map = (user_id.unique().sort()).indices
        user_id = (user_id.view(-1,1)==user_id_map).int().argmax(dim=1)
        user_embedding = local_model.get_embedding(user_id)
        item_embedding = self.get_embedding(item_id)
        rating = (user_embedding*item_embedding).sum(1)
        return rating

class UserEmbedding(RecEmbedding):
    def __init__(self, embedding_size=20):
        super().__init__(dim=1, embedding_size=embedding_size)

def init_local_module(object):
    if 'Client' in object.__class__.__name__:
        object.set_model(UserEmbedding(20))

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.set_model(ItemEmbedding(object.test_data.num_items, 20))


import sys
sys.path.append('../src')
import flgo.algorithm.fedavg as fedavg
import flgo.experiment.analyzer
import os
import torch.multiprocessing as mp

task = './my_task'
gen_config = {
    'benchmark':{'name':'flgo.benchmark.mnist_classification'},
    'partitioner':{'name':'IIDPartitioner', 'para':{'num_clients':100}}
}
analysis_plan = {
    'Selector':{'task':'./my_task', 'header':['fedavg',], },
    'Painter':{'Curve':[{'args':{'x':'communication_round', 'y':'valid_loss'}}]},
    'Table':{'min_value':[{'x':'valid_loss'}]},
}
if __name__ == '__main__':
    # generate federated task if task doesn't exist
    if not os.path.exists(task): flgo.gen_task(gen_config, task_path=task)
    # running fedavg on the specified task
    runner = flgo.init(task, fedavg, {'gpu':[0,],'log_file':True, 'num_steps':5})
    runner.run()
    # visualize the experimental result
    flgo.experiment.analyzer.show(analysis_plan)



import copy
import io
from contextlib import redirect_stdout

import numpy as np
import pycocotools.mask as mask_util
import torch
import utils
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval


class CocoEvaluator:
    def __init__(self, coco_gt, iou_types):
        if not isinstance(iou_types, (list, tuple)):
            raise TypeError(f"This constructor expects iou_types of type list or tuple, instead  got {type(iou_types)}")
        coco_gt = copy.deepcopy(coco_gt)
        self.coco_gt = coco_gt

        self.iou_types = iou_types
        self.coco_eval = {}
        for iou_type in iou_types:
            self.coco_eval[iou_type] = COCOeval(coco_gt, iouType=iou_type)

        self.img_ids = []
        self.eval_imgs = {k: [] for k in iou_types}

    def update(self, predictions):
        img_ids = list(np.unique(list(predictions.keys())))
        self.img_ids.extend(img_ids)

        for iou_type in self.iou_types:
            results = self.prepare(predictions, iou_type)
            with redirect_stdout(io.StringIO()):
                coco_dt = COCO.loadRes(self.coco_gt, results) if results else COCO()
            coco_eval = self.coco_eval[iou_type]

            coco_eval.cocoDt = coco_dt
            coco_eval.params.imgIds = list(img_ids)
            img_ids, eval_imgs = evaluate(coco_eval)

            self.eval_imgs[iou_type].append(eval_imgs)

    def synchronize_between_processes(self):
        for iou_type in self.iou_types:
            self.eval_imgs[iou_type] = np.concatenate(self.eval_imgs[iou_type], 2)
            create_common_coco_eval(self.coco_eval[iou_type], self.img_ids, self.eval_imgs[iou_type])

    def accumulate(self):
        for coco_eval in self.coco_eval.values():
            coco_eval.accumulate()

    def summarize(self):
        for iou_type, coco_eval in self.coco_eval.items():
            print(f"IoU metric: {iou_type}")
            coco_eval.summarize()

    def prepare(self, predictions, iou_type):
        if iou_type == "bbox":
            return self.prepare_for_coco_detection(predictions)
        if iou_type == "segm":
            return self.prepare_for_coco_segmentation(predictions)
        if iou_type == "keypoints":
            return self.prepare_for_coco_keypoint(predictions)
        raise ValueError(f"Unknown iou type {iou_type}")

    def prepare_for_coco_detection(self, predictions):
        coco_results = []
        for original_id, prediction in predictions.items():
            if len(prediction) == 0:
                continue

            boxes = prediction["boxes"]
            boxes = convert_to_xywh(boxes).tolist()
            scores = prediction["scores"].tolist()
            labels = prediction["labels"].tolist()

            coco_results.extend(
                [
                    {
                        "image_id": original_id,
                        "category_id": labels[k],
                        "bbox": box,
                        "score": scores[k],
                    }
                    for k, box in enumerate(boxes)
                ]
            )
        return coco_results

    def prepare_for_coco_segmentation(self, predictions):
        coco_results = []
        for original_id, prediction in predictions.items():
            if len(prediction) == 0:
                continue

            scores = prediction["scores"]
            labels = prediction["labels"]
            masks = prediction["masks"]

            masks = masks > 0.5

            scores = prediction["scores"].tolist()
            labels = prediction["labels"].tolist()

            rles = [
                mask_util.encode(np.array(mask[0, :, :, np.newaxis], dtype=np.uint8, order="F"))[0] for mask in masks
            ]
            for rle in rles:
                rle["counts"] = rle["counts"].decode("utf-8")

            coco_results.extend(
                [
                    {
                        "image_id": original_id,
                        "category_id": labels[k],
                        "segmentation": rle,
                        "score": scores[k],
                    }
                    for k, rle in enumerate(rles)
                ]
            )
        return coco_results

    def prepare_for_coco_keypoint(self, predictions):
        coco_results = []
        for original_id, prediction in predictions.items():
            if len(prediction) == 0:
                continue

            boxes = prediction["boxes"]
            boxes = convert_to_xywh(boxes).tolist()
            scores = prediction["scores"].tolist()
            labels = prediction["labels"].tolist()
            keypoints = prediction["keypoints"]
            keypoints = keypoints.flatten(start_dim=1).tolist()

            coco_results.extend(
                [
                    {
                        "image_id": original_id,
                        "category_id": labels[k],
                        "keypoints": keypoint,
                        "score": scores[k],
                    }
                    for k, keypoint in enumerate(keypoints)
                ]
            )
        return coco_results


def convert_to_xywh(boxes):
    xmin, ymin, xmax, ymax = boxes.unbind(1)
    return torch.stack((xmin, ymin, xmax - xmin, ymax - ymin), dim=1)


def merge(img_ids, eval_imgs):
    all_img_ids = utils.all_gather(img_ids)
    all_eval_imgs = utils.all_gather(eval_imgs)

    merged_img_ids = []
    for p in all_img_ids:
        merged_img_ids.extend(p)

    merged_eval_imgs = []
    for p in all_eval_imgs:
        merged_eval_imgs.append(p)

    merged_img_ids = np.array(merged_img_ids)
    merged_eval_imgs = np.concatenate(merged_eval_imgs, 2)

    # keep only unique (and in sorted order) images
    merged_img_ids, idx = np.unique(merged_img_ids, return_index=True)
    merged_eval_imgs = merged_eval_imgs[..., idx]

    return merged_img_ids, merged_eval_imgs


def create_common_coco_eval(coco_eval, img_ids, eval_imgs):
    img_ids, eval_imgs = merge(img_ids, eval_imgs)
    img_ids = list(img_ids)
    eval_imgs = list(eval_imgs.flatten())

    coco_eval.evalImgs = eval_imgs
    coco_eval.params.imgIds = img_ids
    coco_eval._paramsEval = copy.deepcopy(coco_eval.params)


def evaluate(imgs):
    with redirect_stdout(io.StringIO()):
        imgs.evaluate()
    return imgs.params.imgIds, np.asarray(imgs.evalImgs).reshape(-1, len(imgs.params.areaRng), len(imgs.params.imgIds))


import os
import torchvision
import torch
import torchvision.transforms as T
import flgo.benchmark
transform = T.Compose([
    T.ToTensor(),
    T.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
])
target_transform = T.Compose([
    T.PILToTensor(),
    T.Lambda(lambda x: (x-1).squeeze().type(torch.LongTensor))
])
path = os.path.join(flgo.benchmark.data_root,  'OXFORD-IIIT-PET')
# 定义训练集实例，并命名为train_data
train_data = torchvision.datasets.OxfordIIITPet(root=path, split='trainval', download=True, target_types='segmentation', transform=transform, target_transform=target_transform)
# 定义测试集实例，并命名为test_data
test_data = torchvision.datasets.OxfordIIITPet(root=path, split='test', download=True, target_types='segmentation', transform=transform, target_transform=target_transform)
train_data.num_classes = 3
test_data.num_classes = 3

def get_model():
    model = torchvision.models.segmentation.fcn_resnet50(num_classes=3)
    return model


import math
import torch
from torchtext.datasets import WikiText2
from torchtext.data.utils import get_tokenizer
from torchtext.vocab import build_vocab_from_iterator
from torch import nn, Tensor
from torch.nn import TransformerEncoder, TransformerEncoderLayer

train_iter, val_iter, test_iter = WikiText2()
tokenizer = get_tokenizer('basic_english')
vocab = build_vocab_from_iterator(map(tokenizer, train_iter), specials=['<unk>'])
vocab.set_default_index(vocab['<unk>'])

def apply_transform(x):
    return torch.tensor(vocab(tokenizer(x)), dtype=torch.long)

train_data = train_iter.map(apply_transform)
val_data = val_iter.map(apply_transform)
test_data = test_iter.map(apply_transform)

class PositionalEncoding(nn.Module):

    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    def forward(self, x: Tensor) -> Tensor:
        """
        Arguments:
            x: Tensor, shape ``[seq_len, batch_size, embedding_dim]``
        """
        x = x + self.pe[:x.size(0)]
        return self.dropout(x)

class TransformerModel(nn.Module):

    def __init__(self, ntoken: int, d_model: int, nhead: int, d_hid: int,
                 nlayers: int, dropout: float = 0.5):
        super().__init__()
        self.model_type = 'Transformer'
        self.pos_encoder = PositionalEncoding(d_model, dropout)
        encoder_layers = TransformerEncoderLayer(d_model, nhead, d_hid, dropout)
        self.transformer_encoder = TransformerEncoder(encoder_layers, nlayers)
        self.embedding = nn.Embedding(ntoken, d_model)
        self.d_model = d_model
        self.linear = nn.Linear(d_model, ntoken)

        self.init_weights()

    def init_weights(self) -> None:
        initrange = 0.1
        self.embedding.weight.data.uniform_(-initrange, initrange)
        self.linear.bias.data.zero_()
        self.linear.weight.data.uniform_(-initrange, initrange)

    def forward(self, src: Tensor, src_mask: Tensor = None) -> Tensor:
        """
        Arguments:
            src: Tensor, shape ``[seq_len, batch_size]``
            src_mask: Tensor, shape ``[seq_len, seq_len]``

        Returns:
            output Tensor of shape ``[seq_len, batch_size, ntoken]``
        """
        src = self.embedding(src) * math.sqrt(self.d_model)
        src = self.pos_encoder(src)
        output = self.transformer_encoder(src, src_mask)
        output = self.linear(output)
        return output

def get_model():
    ntokens = len(vocab)  # size of vocabulary
    emsize = 200  # embedding dimension
    d_hid = 200  # dimension of the feedforward network model in ``nn.TransformerEncoder``
    nlayers = 2  # number of ``nn.TransformerEncoderLayer`` in ``nn.TransformerEncoder``
    nhead = 2  # number of heads in ``nn.MultiheadAttention``
    dropout = 0.2
    model = TransformerModel(ntokens, emsize, nhead, d_hid, nlayers, dropout)
    return model


from flgo.utils.fmodule import FModule
import torch

class Model(FModule):
    def __init__(self, dim_in = 2, dim_out = 1):
        super(Model, self).__init__()
        self.x = torch.nn.Parameter(torch.zeros(dim_in).unsqueeze(0))

    def forward(self, e):
        output = e-self.x.repeat((len(e), 1))
        output = torch.diag(torch.mm(output, output.T))
        return output

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import flgo.benchmark.toolkits.partition
import flgo.benchmark.crop_classification.model.cnn as cnn

default_partitioner = flgo.benchmark.toolkits.partition.IIDPartitioner
default_partition_para = {'num_clients':100}
default_model = cnn


r"""
This module is designed for fast creating federated tasks. For example,
in FL, a commonly used benchmark is federated MNIST that splits MNIST
into 100 shards and each shard contains data of two types of labels.

In FLGo, three basic components are created to describe a general
procedure that can easily convert various ML tasks into federated ones.

Components:
    * `TaskGenerator`
        - load the original dataset
        - partition the original dataset into local_movielens_recommendation data

    * `TaskPipe`
        - store the partition information of TaskGenerator into the disk
            when generating federated tasks
        - load the original dataset and the partition information to
            create the federated scenario when optimizing models

    * `TaskCalculator`
        - support task-specific computation when optimizing models, such
            as putting data into device, computing loss, evaluating models,
            and creating the data loader

The architecture of a complete federate benchmark is shown as follows:

```
benchmark_name                  # benchmark folder
├─ core.py                      # core file
│   ├─ TaskGenerator            # class TaskGenerator(...)
│   ├─ TaskPipe                 # class TaskPipe(...)
│   └─ TaskCalculator           # class TaskCalculator(...)
│
├─  model                       # model folder (i.e. contains various types of models)
│   ├─ model1_name.py           # model 1 (e.g. CNN)
│   ├─ ...
│   └─ modelN_name.py           # model N (e.g. ResNet)
│       ├─ init_local_module    # the function initializes personal models for parties
│       └─ init_global_module   # the function initializes the global models for parties
│
└─ __init__.py                  # containing the variable default_model
```

**Example**:
The architecture of MNIST is
```
├─ core.py
│   ├─ TaskGenerator
│   ├─ TaskPipe
│   └─ TaskCalculator
├─  model
│   ├─ cnn.py
│   └─ mlp.py
│       ├─ init_local_module
│       └─ init_global_module
└─ __init__.py
```

The details of implementing a customized benchmark are in Tutorial.3
"""
import os

path = os.path.abspath(os.path.dirname(__file__))
data_root = os.path.join(path, 'RAW_DATA')


import torch_geometric.data
from flgo.benchmark.toolkits import BasicTaskGenerator, BasicTaskCalculator, BasicTaskPipe
import os.path
import urllib
import torch
from torch.utils.data import DataLoader
import os
import zipfile
from urllib.request import urlopen
import numpy as np
import flgo
import scipy.io as scio
import pandas as pd
try:
    import ujson as json
except:
    import json

def download_from_url(url= None, filepath = '.'):
    """Download dataset from url to filepath."""
    if url: urllib.request.urlretrieve(url, filepath)
    return filepath

def extract_from_zip(src_path, target_path):
    """Unzip the .zip file (src_path) to target_path"""
    f = zipfile.ZipFile(src_path)
    f.extractall(target_path)
    targets = f.namelist()
    f.close()
    return [os.path.join(target_path, tar) for tar in targets]

class TaskGenerator(BasicTaskGenerator):
    _DATASET_INFO = {
        'url':"https://www.cse.msu.edu/~tangjili/datasetcode/epinions.zip",
        'link_type':'trust',
        'zipfile': 'epinions.zip',
    }
    # rawdata_path=os.path.join(flgo.benchmark.data_root, 'CIAODVD')
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.data_root, 'EPINIONS'), scale = 1.0, order='max', min_items = 3):
        super(TaskGenerator, self).__init__(benchmark='social_splitted_epinions', rawdata_path=rawdata_path)
        self.scene='vertical'
        for key,value in self._DATASET_INFO.items(): setattr(self, key, value)
        self.scale = scale
        self.order = order
        self.min_items = min_items
        self.num_users = 0

    def partition(self, *args, **kwargs):
        # sort and filter users
        all_users = self.ratings['userID'].unique()
        self.num_users = int(len(all_users)*self.scale)
        tmp = self.ratings['userID'].value_counts()
        tmp = tmp[tmp.values>=self.min_items]
        sorted_users = list(tmp.index)
        # user_size = tmp.values
        self.num_users = min(len(sorted_users), self.num_users)
        if self.order == 'max':
            users = sorted_users[:self.num_users]
        elif self.order == 'min':
            sorted_users.reverse()
            users = sorted_users[:self.num_users]
        else:
            users = list(np.random.choice(sorted_users, self.num_users, replace=False))
        # collect items according to the users
        ratings = self.ratings[self.ratings['userID'].isin(users)]
        items = ratings['movieID'].unique()
        links = self.links[(self.links['trustorID'].isin(users)) & (self.links['trusteeID'].isin(users))]
        user_map = {uid:uidx for uidx,uid in enumerate(users)}
        item_map = {vid:vidx for vidx,vid in enumerate(items)}
        ratings['userID'] = ratings['userID'].map(user_map)
        ratings['movieID'] = ratings['movieID'].map(item_map)
        links['trustorID'] = links['trustorID'].map(user_map)
        links['trusteeID'] = links['trusteeID'].map(user_map)
        self.users = list(range(len(users)))
        self.items = list(range(len(items)))
        self.ratings = ratings
        self.links = links

    def get_task_name(self):
        return 'social_splitted_epinions_u{}_o{}_m{}'.format(self.num_users, self.order,self.min_items)

    def load_data(self, *args, **kwargs):
        # download zip file
        zippath = os.path.join(self.rawdata_path, self.zipfile)
        if not os.path.exists(zippath):
            if not os.path.exists(self.rawdata_path):
                os.makedirs(self.rawdata_path)
            _ = download_from_url(self.url, zippath)
        # extract zip
        if not os.path.exists(os.path.join(self.rawdata_path, 'epinions')): _ = extract_from_zip(zippath, self.rawdata_path)
        # read data into memory
        ratings = scio.loadmat(os.path.join(self.rawdata_path, 'epinions', 'rating.mat'))['rating'][:,[0,1,3]]
        ratings = ratings.T
        rating_dict = {'userID':ratings[0], 'movieID':ratings[1], 'Rating':ratings[2]}
        self.ratings = pd.DataFrame(rating_dict)
        links = scio.loadmat(os.path.join(self.rawdata_path, 'epinions', 'trustnetwork.mat'))['trustnetwork']
        links = links.T
        link_dict = {'trustorID':links[0],'trusteeID':links[1]}
        self.links = pd.DataFrame(link_dict)
        return

class TaskPipe(BasicTaskPipe):
    class TaskDataset(torch.utils.data.Dataset):
        def __init__(self, data):
            self.user_ids = torch.LongTensor([d[0] for d in data])
            self.movie_ids = torch.LongTensor([d[1] for d in data])
            self.ratings = torch.FloatTensor([d[2] for d in data])

        def __getitem__(self, item):
            return self.user_ids[item], self.movie_ids[item], self.ratings[item]

        def __len__(self):
            return len(self.user_ids)

    def save_task(self, generator):
        party_names = ['Advertiser', 'MeidaCompany']
        feddata = {'party_names': party_names, 'rawdata_path': generator.rawdata_path}
        feddata['Advertiser'] = {
            'data':{
                'userID':generator.users,
                'itemID':generator.items,
                'ratings': generator.ratings.values.tolist(),
                'with_label':True
            }
        }
        feddata['MeidaCompany'] = {
            'data': {
                'userID': generator.users,
                'links': generator.links.values.tolist(),
                'with_label':False
            }
        }
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)

    def load_data(self, running_time_option, *args, **kwargs):
        task_data = {}
        for pid, party_name in enumerate(self.feddata['party_names']):
            pdata = self.feddata[party_name]['data']
            if party_name=='Advertiser':
                ratings = pdata['ratings']
                users = pdata['userID']
                tmp = pd.DataFrame(ratings)
                train_data = []
                test_data = []
                val_data = []
                val_ratio = running_time_option['train_holdout']
                for uid in users:
                    udata = tmp[tmp[0]==uid].values.tolist()
                    local_data_size = max(int(len(udata)*0.9), 1)
                    ulocal = udata[:local_data_size]
                    utest = udata[local_data_size:]
                    train_size = int(len(ulocal)*(1-val_ratio))
                    utrain = ulocal[:train_size]
                    uvalid = ulocal[train_size:]
                    train_data.extend(utrain)
                    val_data.extend(uvalid)
                    test_data.extend(utest)
                train_data = self.TaskDataset(train_data)
                val_data = self.TaskDataset(val_data) if len(val_data)>0 else None
                test_data = self.TaskDataset(test_data)
                task_data[party_name] = {'userID': pdata['userID'], 'itemID':pdata['itemID'], 'train':train_data, 'val':val_data, 'test':test_data}
            else:
                social_links = pdata['links']
                edge_index = torch.tensor([[l[0] for l in social_links], [l[1] for l in social_links]], dtype=torch.int64)
                social_data = torch_geometric.data.Data(edge_index=edge_index)
                task_data[party_name] = {'userID':pdata['userID'],'social':social_data}
        return task_data

class TaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super().__init__(device, optimizer_name)
        self.criterion = torch.nn.MSELoss()
        self.DataLoader = DataLoader

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device), data[2].to(self.device)
#
    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False, drop_last=False, *args, **kwargs):
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)

    def compute_loss(self, model, data, *args, **kwargs):
        tdata = self.to_device(data)
        model = model.to(self.device)
        output = model(tdata)
        loss =  self.criterion(output, torch.tensor(tdata[-1], dtype=torch.float))
        return {'loss':loss}

    def test(self, model, dataset, batch_size=64, num_workers = 0, *args, **kwargs):
        model.to(self.device)
        with torch.no_grad():
            model.eval()
            if batch_size == -1: batch_size = len(dataset)
            data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers)
            mse_loss = 0
            mae_loss = 0
            for batch_id, batch_data in enumerate(data_loader):
                batch_data = self.to_device(batch_data)
                outputs = model(batch_data)
                batch_mae = torch.abs(outputs-batch_data[-1]).sum().item()
                batch_mse = torch.nn.MSELoss()(outputs, batch_data[-1]).item()
                mse_loss += batch_mse * len(batch_data[-1])
                mae_loss += batch_mae
            mae = mae_loss/len(dataset)
            rmse = torch.sqrt(torch.tensor(mse_loss/len(dataset)))
        return {'mae': mae, 'rmse': rmse.item()}


"""
train_data (torch.utils.data.Dataset),
test_data (torch.utils.data.Dataset),
and the model (torch.nn.Module) should be implemented here.

"""
import torch.nn as nn
import torch.utils.data
from typing import *
train_data = []
val_data = None
test_data = None

def data_to_device(batch_data, device):
    raise NotImplementedError

def eval(model: nn.Module, data_loader: torch.utils.data.Dataset, device) -> dict:
    raise NotImplementedError

def compute_loss(batch_data, model:nn.Module, device) -> dict:
    raise NotImplementedError

def get_model(*args, **kwargs) -> torch.nn.Module:
    raise NotImplementedError


"""
This is a non-official implementation of MOON proposed in 'Model-Contrastive
Federated Learning (https://arxiv.org/abs/2103.16257)'. The official implementation is in https://github.com/QinbinLi/MOON.
********************************************Note***********************************************
The model used by this algorithm should be formulated by two submodules: encoder and head
"""
from flgo.algorithm.fedbase import BasicServer, BasicClient
import copy
import torch
import torch.nn.functional as F
import flgo.utils.fmodule as fmodule

def model_contrastive_loss(z, z_glob, z_prev, temperature=0.5):
    pos_sim = F.cosine_similarity(z, z_glob, dim=-1)
    logits = pos_sim.reshape(-1, 1)
    if z_prev is not None:
        neg_sim = F.cosine_similarity(z, z_prev, dim=-1)
        # neg_sim = self.cos(z, z_prev)
        logits = torch.cat((logits, neg_sim.reshape(-1, 1)), dim=1)
    logits /= temperature
    return F.cross_entropy(logits, torch.zeros(z.size(0)).long().to(logits.device))

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'mu': 0.1, 'tau':0.5})
        self.output_layer = ".".join([f'[{m}]' if m.isdigit() else f'{m}' for m in list(self.model.state_dict().keys())[-1].split('.')[:-1]])

class Client(BasicClient):
    def initialize(self, *args, **kwargs):
        self.local_model = None
        self.output_layer = self.server.output_layer

    @fmodule.with_multi_gpus
    def train(self, model):
        # init global model and local model
        global_model = copy.deepcopy(model)
        global_model.freeze_grad()
        if self.local_model is not None:
            self.local_model.to(model.get_device())

        feature_maps = []
        def hook(model, input, output):
            feature_maps.append(input)
        eval('global_model.{}'.format(self.output_layer)).register_forward_hook(hook)
        if self.local_model is not None: eval("self.local_model.{}".format(self.output_layer)).register_forward_hook(hook)
        eval('model.{}'.format(self.output_layer)).register_forward_hook(hook)

        model.train()
        optimizer = self.calculator.get_optimizer(model, lr=self.learning_rate, weight_decay=self.weight_decay, momentum=self.momentum)
        for iter in range(self.num_steps):
            batch_data = self.get_batch_data()
            batch_data = self.calculator.to_device(batch_data)
            model.zero_grad()
            loss = self.calculator.compute_loss(model, batch_data)['loss']
            _ = global_model(batch_data[0])
            if self.local_model is not None:
                _ = self.local_model(batch_data[0])
            else:
                feature_maps.append((None,))
            z_prev = feature_maps.pop()[0]
            z_glob = feature_maps.pop()[0]
            z = feature_maps.pop()[0]
            loss_con = model_contrastive_loss(z, z_glob, z_prev, self.tau)
            loss = loss + self.mu * loss_con
            loss.backward()
            if self.clip_grad>0:torch.nn.utils.clip_grad_norm_(parameters=model.parameters(), max_norm=self.clip_grad)
            optimizer.step()

        self.local_model = copy.deepcopy(model).to(torch.device('cpu'))
        self.local_model.freeze_grad()
        return


"""This is a non-official implementation of 'Fair Resource Allocation in
Federated Learning' (http://arxiv.org/abs/1905.10497).  And this implementation
refers to the official github repository https://github.com/litian96/fair_flearn """
import flgo.algorithm.fedbase as fedbase
import flgo.utils.fmodule as fmodule
import copy

class Server(fedbase.BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'q': 1.0})

    def iterate(self):
        self.selected_clients = self.sample()
        res = self.communicate(self.selected_clients)
        self.model = self.model - fmodule._model_sum(res['dk']) / sum(res['hk'])
        return len(self.received_clients) > 0

class Client(fedbase.BasicClient):
    def unpack(self, package):
        model = package['model']
        self.global_model = copy.deepcopy(model)
        return model

    def pack(self, model):
        Fk = self.test(self.global_model, 'train')['loss'] + 1e-8
        L = 1.0 / self.learning_rate
        delta_wk = L * (self.global_model - model)
        dk = (Fk ** self.q) * delta_wk
        hk = self.q * (Fk ** (self.q - 1)) * (delta_wk.norm() ** 2) + L * (Fk ** self.q)
        self.global_model = None
        return {'dk': dk, 'hk': hk}



"""
This is a non-official implementation of the work 'Secure Federated Matrix Factorization' (https://arxiv.org/abs/1906.05108).
"""
import collections
import torch
import copy
from .fedbase import BasicServer
from .fedbase import BasicClient
from phe import paillier
import numpy as np
from tqdm import tqdm

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        # hyper-parameters
        self.init_algo_para({'lambda_l2':1e-4, 'part':1})
        self.gv.public_key, self.gv.private_key = paillier.generate_paillier_keypair(n_length=1024, )
        # initialize item vectors as [0.01 ... 0.01] as the official code did (just for testing)
        # encrypt the item vectors
        if self.part < 0:
            # if part < 0, it won't encrypt item embeddings
            self.encrypted_item_vectors = self.model.get_embedding().detach().cpu().numpy()
        else:
            tmp = self.model.get_embedding().detach().cpu().numpy().tolist()
            self.encrypted_item_vectors = []
            self.gv.logger.info('Encrypt all the item embeddings...')
            for vector in tqdm(tmp):
                self.encrypted_item_vectors.append([self.gv.public_key.encrypt(e, precision=1e-5) for e in vector])

    def iterate(self):
        self.selected_clients = self.sample()
        en_grads = self.communicate(self.selected_clients)['encrypted_gradient']
        self.aggregate(en_grads)

        # update decrypted item embeddings

    def pack(self, *args, **kwargs):
        return {
            'model': copy.deepcopy(self.model),
            'encrypted_item_vectors': self.encrypted_item_vectors
        }

    def aggregate(self, encrypted_grads):
        if self.part<0:
            for g in encrypted_grads:
                self.encrypted_item_vectors -= g
        elif self.part==0:
                for eng in encrypted_grads:
                    for i in range(len(self.encrypted_item_vectors)):
                        for j in range(len(self.encrypted_item_vectors[i])):
                            self.encrypted_item_vectors[i][j] = self.encrypted_item_vectors[i][j] - eng[i][j]*1.0/len(encrypted_grads)
        elif self.part==1:
            for eng in encrypted_grads:
                for item_id in eng:
                    for j in range(len(self.encrypted_item_vectors[item_id])):
                        self.encrypted_item_vectors[item_id][j] = self.encrypted_item_vectors[item_id][j] - eng[item_id][j]*1.0/len(encrypted_grads)

    def global_test(self, model=None, flag:str='val'):
        if model is None: model=self.model
        all_metrics = collections.defaultdict(list)
        for c in self.clients:
            client_metrics = c.test(model, flag)
            for met_name, met_val in client_metrics.items():
                all_metrics[met_name].append(met_val)
        return all_metrics

    def test(self, model=None, flag='test'):
        data = self.test_data if flag=='test' else self.val_data
        if data is None: return {}
        # construct user embeddings
        clients = sorted(self.clients, key=lambda x: x.train_data.user_id[0])
        user_embeddings = np.array([c.model.get_embedding().view(-1).detach().cpu().numpy() for c in clients])
        local_model = self.model.__class__(len(clients), 20).to(self.device)
        local_model.set_embedding(user_embeddings)
        # construct item embeddings
        if self.part<0:
            tmp = self.encrypted_item_vectors
        else:
            tmp = np.array([[self.gv.private_key.decrypt(e) for e in vector] for vector in tqdm(self.encrypted_item_vectors)], dtype=np.float32)
        self.model.set_embedding(tmp)
        return self.calculator.test(models = (self.model, local_model), dataset=data, batch_size=self.option['test_batch_size'], num_workers=self.option['num_workers'], pin_memory=self.option['pin_memory'])

class Client(BasicClient):
    def reply(self, svr_pkg):
        model = self.unpack(svr_pkg)
        grad = self.train(model)
        cpkg = self.pack(grad)
        return cpkg

    def unpack(self, received_pkg):
        """Return the decrypted item embeddings (i.e. the instance of ItemEmbedding())"""
        encrypted_item_vectors = received_pkg['encrypted_item_vectors']
        model = received_pkg['model']
        # local_movielens_recommendation decrypt
        if self.part<0:
            item_embeddings = encrypted_item_vectors
        else:
            item_vectors_np = []
            self.gv.logger.info('Client {} decrypt the item embeddings...'.format(self.id))
            for vector in tqdm(encrypted_item_vectors):
                item_vectors_np.append([self.gv.private_key.decrypt(e) for e in vector])
            item_embeddings = np.array(item_vectors_np)
        model.set_embedding(item_embeddings)
        return model

    def train(self, global_model):
        # self.gv.logger.info('Client {} locally train the models...'.format(self.id))
        # zero grad
        original_model = copy.deepcopy(global_model)
        global_model.zero_grad()
        self.model.zero_grad()
        # get a batch of data
        optimizer = torch.optim.SGD([{'params':self.model.parameters(), 'lr':self.learning_rate}, {'params':global_model.parameters(), 'lr':self.learning_rate}], weight_decay=self.lambda_l2)
        for _ in range(self.num_steps):
            batch_data = self.get_batch_data()
            self.model.to(self.device)
            global_model.to(self.device)
            loss = self.calculator.compute_loss((global_model, self.model), batch_data)['loss']
            # predict = global_model(batch_data, self.model.get_embedding())
            # loss = ((batch_data['rating']-predict)**2).sum()/len(batch_data['item_id'])
            loss.backward()
            if self.clip_grad>0:torch.nn.utils.clip_grad_norm_(parameters=self.model.parameters(), max_norm=self.clip_grad)
            if self.clip_grad>0:torch.nn.utils.clip_grad_norm_(parameters=global_model.parameters(), max_norm=self.clip_grad)
            optimizer.step()
        global_model.to(torch.device('cpu'))
        gradient = (global_model.get_embedding()-original_model.get_embedding()).detach().cpu().numpy()
        return gradient

    def pack(self, gradient):
        if self.part==0:
            encrypted_gradient = []
            for vector in tqdm(gradient):
                encrypted_gradient.append([self.gv.public_key.encrypt(float(e), precision=1e-5) for e in vector])
        elif self.part==1:
            self.gv.logger.info('Client {} encrypt the gradient...'.format(self.id))
            encrypted_gradient = {}
            item_list = list(set(self.train_data.item_id.tolist()))
            for item_id in tqdm(item_list):
                encrypted_gradient[item_id] = [self.gv.public_key.encrypt(float(e), precision=1e-5) for e in gradient[item_id]]
        else:
            encrypted_gradient = gradient
        return {'encrypted_gradient': encrypted_gradient}

    def test(self, global_model, flag='val'):
        data = self.train_data if flag=='train' else self.val_data
        if data is None: return {}
        return self.calculator.test(models=(global_model, self.model), dataset=data, batch_size=self.option['test_batch_size'], pin_memory=self.option['pin_memory'], num_workers=self.option['num_workers'])




import sys
from typing import Any

sys.path.append('/home/wz/anaconda3/lib/python3.9/site-packages/')
import os
import torch
from torchvision.datasets.utils import download_and_extract_archive
import flgo
import torchvision
from flgo.benchmark.toolkits.cv.detection import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator
from ..toolkits.cv.detection.utils import get_transform
from ..toolkits.cv.detection.coco_utils import ConvertCocoPolysToMask
from ..toolkits.cv.detection.transforms import Compose

CAT_LIST = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80]
PATHS = {
        "train": ("train2017", os.path.join("annotations", "instances_train2017.json")),
        "val": ("val2017", os.path.join("annotations", "instances_val2017.json")),
}

class CocoDetection(torchvision.datasets.CocoDetection):
    def __init__(self, root:str, annFile:str, transforms:Any):
        super().__init__(root, annFile)
        self._transforms = transforms

    def __getitem__(self, idx):
        img, target = super().__getitem__(idx)
        image_id = self.ids[idx]
        target = dict(image_id=image_id, annotations=target)
        if self._transforms is not None:
            img, target = self._transforms(img, target)
        return img, target

builtin_class = CocoDetection
train_transform = Compose([ConvertCocoPolysToMask(), get_transform(train=True)])
test_transform = Compose([ConvertCocoPolysToMask(), get_transform(train=False)])
path = os.path.join(flgo.benchmark.data_root,  'COCO')

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=path):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1], rawdata_path=rawdata_path,
                                            builtin_class=builtin_class)
        self.num_classes = 91
        self.test_transform = test_transform
        self.train_transform = train_transform
        self.additional_option = {}
        self.train_image_path = os.path.join(self.rawdata_path, "train2017")
        self.train_annotation_path = os.path.join(self.rawdata_path,"annotations", "instances_train2017.json")
        self.val_image_path = os.path.join(self.rawdata_path, "val2017")
        self.val_annotation_path = os.path.join(self.rawdata_path,"annotations", "instances_val2017.json")
        self.train_additional_option = {'root':self.train_image_path,'image_set':'train', 'annFile':self.train_annotation_path}
        self.test_additional_option = {'root':self.val_image_path,'image_set':'val', 'annFile':self.val_annotation_path}

    def load_data(self):
        # load the datasets
        train_default_init_para = {'transforms':self.train_transform}
        test_default_init_para = {'transforms':self.test_transform}
        train_default_init_para.update(self.additional_option)
        train_default_init_para.update(self.train_additional_option)
        test_default_init_para.update(self.additional_option)
        test_default_init_para.update(self.test_additional_option)
        if 'kwargs' not in self.builtin_class.__init__.__annotations__:
            train_pop_key = [k for k in train_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
            test_pop_key = [k for k in test_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
            for k in train_pop_key: train_default_init_para.pop(k)
            for k in test_pop_key: test_default_init_para.pop(k)
        # init datasets
        self.train_data = self.builtin_class(**train_default_init_para)
        self.test_data = self.builtin_class(**test_default_init_para)

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, builtin_class)
        self.test_transform = test_transform
        self.train_transform = train_transform

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        train_default_init_para = {'transforms':self.train_transform}
        test_default_init_para = {'transforms':self.test_transform}
        if 'additional_option' in self.feddata.keys():
            train_default_init_para.update(self.feddata['additional_option'])
            test_default_init_para.update(self.feddata['additional_option'])
        if 'train_additional_option' in self.feddata.keys(): train_default_init_para.update(self.feddata['train_additional_option'])
        if 'test_additional_option' in self.feddata.keys(): test_default_init_para.update(self.feddata['test_additional_option'])
        if 'kwargs' not in self.builtin_class.__init__.__annotations__:
            train_pop_key = [k for k in train_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
            test_pop_key = [k for k in test_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
            for k in train_pop_key: train_default_init_para.pop(k)
            for k in test_pop_key: test_default_init_para.pop(k)
        train_data = self.builtin_class(**train_default_init_para)
        test_data = self.builtin_class(**test_default_init_para)
        test_data = self.TaskDataset(test_data, list(range(len(test_data))), None )
        # rearrange data for server
        server_data_test, server_data_val = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'val': server_data_val}}
        # rearrange data for clients
        local_perturbation = self.feddata['local_perturbation'] if 'local_perturbation' in self.feddata.keys() else [None for _ in self.feddata['client_names']]
        for cid, cname in enumerate(self.feddata['client_names']):
            cpert = None if  local_perturbation[cid] is None else [torch.tensor(t) for t in local_perturbation[cid]]
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'], cpert )
            cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
            if running_time_option['train_holdout']>0 and running_time_option['local_test']:
                cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
            else:
                cdata_test = None
            task_data[cname] = {'train':cdata_train, 'val':cdata_val, 'test': cdata_test}
        return task_data

TaskCalculator = GeneralCalculator


from flgo.experiment.logger import BasicLogger
import numpy as np

class GValLogger(BasicLogger):
    def log_once(self, *args, **kwargs):
        val_metric = self.server.test(flag='val')
        for met_name, met_val in val_metric.items():
            self.output['val_' + met_name].append(met_val)
        self.show_current_output()


from flgo.experiment.logger import BasicLogger
import numpy as np
import flgo.simulator.base as ss

class SimpleLogger(BasicLogger):
    r"""Simple Logger. Only evaluating model performance on testing dataset and validation dataset."""
    def initialize(self):
        """This method is used to record the stastic variables that won't change across rounds (e.g. local_movielens_recommendation data size)"""
        for c in self.participants:
            self.output['client_datavol'].append(len(c.train_data))

    """This logger only records metrics on validation dataset"""
    def log_once(self, *args, **kwargs):
        self.info('Current_time:{}'.format(self.clock.current_time))
        self.output['time'].append(self.clock.current_time)
        test_metric = self.coordinator.test()
        for met_name, met_val in test_metric.items():
            self.output['test_' + met_name].append(met_val)
        val_metrics = self.coordinator.global_test(flag='val')
        local_data_vols = [c.datavol for c in self.participants]
        total_data_vol = sum(local_data_vols)
        for met_name, met_val in val_metrics.items():
            self.output['val_'+met_name+'_dist'].append(met_val)
            self.output['val_' + met_name].append(1.0 * sum([client_vol * client_met for client_vol, client_met in zip(local_data_vols, met_val)]) / total_data_vol)
            self.output['mean_val_' + met_name].append(np.mean(met_val))
            self.output['std_val_' + met_name].append(np.std(met_val))
        # output to stdout
        self.show_current_output()


import copy

import torch
from tqdm import trange
import flgo.algorithm.fedbase as fab

class Server(fab.BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'tune_key':'loss', 'larger_is_better': False})

    def run(self):
        self.gv.logger.time_start('Total Time Cost')
        # evaluating initial model performance
        self.gv.logger.info("--------------Initial Evaluation--------------")
        self.gv.logger.time_start('Eval Time Cost')
        self.gv.logger.log_once()
        self.gv.logger.time_end('Eval Time Cost')
        # Standalone Training
        for c in self.clients: c.finetune()
        self.gv.logger.info("=================End==================")
        self.gv.logger.time_end('Total Time Cost')
        self.gv.logger.time_start('Eval Time Cost')
        self.gv.logger.log_once()
        self.gv.logger.time_end('Eval Time Cost')
        # save results as .json file
        self.gv.logger.save_output_as_json()
        return

class Client(fab.BasicClient):
    def initialize(self, *args, **kwargs):
        self.model = copy.deepcopy(self.server.model)

    def finetune(self):
        dataloader = self.calculator.get_dataloader(self.train_data, batch_size=self.batch_size)
        optimizer = self.calculator.get_optimizer(self.model, lr=self.learning_rate, weight_decay=self.weight_decay, momentum=self.momentum)
        epoch_iter = trange(self.num_epochs+1)
        op_model_dict = copy.deepcopy(self.model.state_dict())
        op_met = self.test(self.model, 'val')
        op_epoch = 0
        for e in epoch_iter:
            if e<self.num_epochs:
                for batch_id, batch_data in enumerate(dataloader):
                    optimizer.zero_grad()
                    batch_data = self.calculator.to_device(batch_data)
                    loss = self.calculator.compute_loss(self.model, batch_data)['loss']
                    loss.backward()
                    if self.clip_grad > 0: torch.nn.utils.clip_grad_norm_(parameters=self.model.parameters(), max_norm=self.clip_grad)
                    optimizer.step()
                val_metric = self.test(self.model, 'val')
                if (self.larger_is_better and val_metric[self.tune_key]>op_met[self.tune_key]) or ((not self.larger_is_better) and val_metric[self.tune_key]<op_met[self.tune_key]):
                    op_met = val_metric
                    op_epoch = e+1
                    op_model_dict = copy.deepcopy(self.model.state_dict())
                terms = ["Client {}".format(self.id), "Epoch {}/{}".format(e+1, self.num_epochs)]
                terms.extend(['val_{}: {:.4f}'.format(k,v) for k,v in val_metric.items()])
                epoch_iter.set_description("\t".join(terms))
            else:
                terms = ["Client {}".format(self.id), "Optimal Epoch {}/{}".format(op_epoch, self.num_epochs)]
                terms.extend(['val_{}: {:.4f}'.format(k, v) for k, v in op_met.items()])
                epoch_iter.set_description("\t".join(terms))
                self.model.load_state_dict(op_model_dict)
        return


